import contextlib
import random
import re
import socket
import time
from collections.abc import AsyncGenerator, Iterable, Iterator
from typing import (
    Any,
    Dict,
    List,
    Set,
    Tuple,
    cast,
    TYPE_CHECKING,
    Generic,
    TypeVar,
)

from django.conf import settings
from django.core.cache.backends.base import DEFAULT_TIMEOUT, get_key_func
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from valkey.exceptions import (
    ConnectionError as ValkeyConnectionError,
    ResponseError,
    TimeoutError as ValkeyTimeoutError,
)
from valkey.typing import AbsExpiryT, EncodableT, ExpiryT, KeyT, PatternT

from django_valkey import pool
from django_valkey.base import ATTR_DOES_NOT_EXIST
from django_valkey.compressors.identity import IdentityCompressor
from django_valkey.exceptions import ConnectionInterrupted
from django_valkey.serializers.pickle import PickleSerializer
from django_valkey.util import CacheKey, decode, encode, make_key, make_pattern

if TYPE_CHECKING:
    from valkey.lock import Lock
    from valkey.asyncio.lock import Lock as AsyncLock
    from django_valkey.cache import ValkeyCache


_main_exceptions = (
    ValkeyTimeoutError,
    ResponseError,
    ValkeyConnectionError,
    socket.timeout,
)

special_re = re.compile("([*?[])")


def glob_escape(s: str) -> str:
    return special_re.sub(r"[\1]", s)


Backend = TypeVar("Backend")


class BaseClient(Generic[Backend]):
    def __init__(
        self,
        server: str | Iterable,
        params: Dict[str, Any],
        backend: "ValkeyCache",
    ) -> None:
        self._backend = backend
        self._server = server
        if not self._server:
            error_message = "Missing connections string"
            raise ImproperlyConfigured(error_message)
        if not isinstance(self._server, (list, tuple, set)):
            self._server = self._server.split(",")

        self._params = params

        self.reverse_key = get_key_func(
            params.get("REVERSE_KEY_FUNCTION")
            or "django_valkey.util.default_reverse_key"
        )

        self._clients: List[Backend | Any | None] = [None] * len(self._server)
        self._options: dict = params.get("OPTIONS", {})
        self._replica_read_only = self._options.get("REPLICA_READ_ONLY", True)

        serializer_path = self._options.get(
            "SERIALIZER", "django_valkey.serializers.pickle.PickleSerializer"
        )
        serializer_cls = import_string(serializer_path)

        compressor_path = self._options.get(
            "COMPRESSOR", "django_valkey.compressors.identity.IdentityCompressor"
        )
        compressor_cls = import_string(compressor_path)

        self._serializer: PickleSerializer | Any = serializer_cls(options=self._options)
        self._compressor: IdentityCompressor | Any = compressor_cls(
            options=self._options
        )

        self._connection_factory = getattr(
            settings, "DJANGO_VALKEY_CONNECTION_FACTORY", self.CONNECTION_FACTORY_PATH
        )
        self.connection_factory = pool.get_connection_factory(
            options=self._options, path=self._connection_factory
        )

    def _has_compression_enabled(self) -> bool:
        return (
            self._options.get(
                "COMPRESSOR", "django_valkey.compressors.identity.IdentityCompressor"
            )
            != "django_valkey.compressors.identity.IdentityCompressor"
        )

    def get_next_client_index(
        self, write: bool = True, tried: List[int] | None = None
    ) -> int:
        """
        Return a next index for read client. This function implements a default
        behavior for get a next read client for a replication setup.

        Overwrite this function if you want a specific
        behavior.
        """
        if write or len(self._server) == 1:
            return 0

        if tried is None:
            tried = []

        if tried and len(tried) < len(self._server):
            not_tried = [i for i in range(0, len(self._server)) if i not in tried]
            return random.choice(not_tried)

        return random.randint(1, len(self._server) - 1)

    def decode(self, value: bytes) -> Any:
        """
        Decode the given value.
        """
        return decode(value, serializer=self._serializer, compressor=self._compressor)

    def encode(self, value: EncodableT) -> bytes | int | float:
        """
        Encode the given value.
        """

        return encode(
            value=value, serializer=self._serializer, compressor=self._compressor
        )

    def _decode_iterable_result(
        self, result: Any, convert_to_set: bool = True
    ) -> List[Any] | Any | None:
        if result is None:
            return None
        if isinstance(result, list):
            if convert_to_set:
                return {self.decode(value) for value in result}
            return [self.decode(value) for value in result]
        return self.decode(result)

    def make_key(
        self, key: KeyT, version: int | None = None, prefix: str | None = None
    ) -> KeyT:
        """Return key as a CacheKey instance so it has additional methods"""
        return make_key(
            key,
            key_func=self._backend.key_func,
            version=version or self._backend.version,
            prefix=prefix or self._backend.key_prefix,
        )

    def make_pattern(
        self, pattern: str, version: int | None = None, prefix: str | None = None
    ) -> KeyT:
        return make_pattern(
            pattern=pattern,
            key_func=self._backend.key_func,
            version=version or self._backend.version,
            prefix=prefix or self._backend.key_prefix,
        )


class ClientCommands(Generic[Backend]):
    def __contains__(self, key: KeyT) -> bool:
        return self.has_key(key)

    def _get_client(self, write=True, tried=None, client=None, **kwargs):
        if client:
            return client
        return self.get_client(write=write, tried=tried, **kwargs)

    def get_client(
        self: BaseClient,
        write: bool = True,
        tried: List[int] | None = None,
        **kwargs,
    ) -> Backend | Any:
        """
        Method used for obtain a raw valkey client.

        This function is used by almost all cache backend
        operations for obtain a native valkey client/connection
        instance.
        """
        index = self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = self.connect(index)

        return self._clients[index]  # type:ignore

    def get_client_with_index(
        self: BaseClient,
        write: bool = True,
        tried: List[int] | None = None,
    ) -> Tuple[Backend | Any, int]:
        """
        Method used for obtain a raw valkey client.

        This function is used by almost all cache backend
        operations for obtain a native valkey client/connection
        instance.
        """
        index = self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = self.connect(index)

        return self._clients[index], index  # type:ignore

    def connect(self: BaseClient, index: int = 0) -> Backend | Any:
        """
        Given a connection index, returns a new raw valkey client/connection
        instance. Index is used for replication setups and indicates that
        connection string should be used. In normal setups, index is 0.
        """
        return self.connection_factory.connect(self._server[index])

    def disconnect(
        self: BaseClient, index: int = 0, client: Backend | Any | None = None
    ) -> None:
        """
        delegates the connection factory to disconnect the client
        """
        if client is None:
            client = self._clients[index]

        if client is not None:
            self.connection_factory.disconnect(client)

    def set(
        self: BaseClient,
        key: KeyT,
        value: EncodableT,
        timeout: int | float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Backend | Any | None = None,
        nx: bool = False,
        xx: bool = False,
    ) -> bool:
        """
        Persist a value to the cache, and set an optional expiration time.

        Also supports optional nx parameter. If set to True - will use valkey
        setnx instead of set.
        """
        nkey = self.make_key(key, version=version)
        nvalue = self.encode(value)

        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        original_client = client
        tried: List[int] = []
        while True:
            try:
                if client is None:
                    client, index = self.get_client_with_index(write=True, tried=tried)

                if timeout is not None:
                    # Convert to milliseconds
                    timeout = int(timeout * 1000)

                    if timeout <= 0:
                        if nx:
                            # Using negative timeouts when nx is True should
                            # not expire (in our case delete) the value if it exists.
                            # Obviously expire not existent value is noop.
                            return not self.has_key(key, version=version, client=client)

                        # TODO: check if this is still valid
                        # valkey doesn't support negative timeouts in ex flags
                        # so it seems that it's better to just delete the key
                        # than to set it and then expire in a pipeline
                        return bool(self.delete(key, client=client, version=version))

                return client.set(nkey, nvalue, nx=nx, px=timeout, xx=xx)
            except _main_exceptions as e:
                if (
                    not original_client
                    and not self._replica_read_only
                    and len(tried) < len(self._server)
                ):
                    tried.append(index)
                    client = None
                    continue
                raise ConnectionInterrupted(connection=client) from e

    def incr_version(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Adds delta to the cache version for the supplied key. Returns the
        new version.
        """

        client = self._get_client(write=True, client=client)

        new_key, old_key, value, ttl, version = self._incr_version(
            key, delta, version, client
        )
        self.set(new_key, value, timeout=ttl, client=client)
        self.delete(old_key, client=client)
        return version + delta

    def decr_version(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        return self.incr_version(key=key, delta=-delta, version=version, client=client)

    def _incr_version(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Tuple:
        if version is None:
            version = self._backend.version

        old_key = self.make_key(key, version)
        value = self.get(old_key, version=version, client=client)

        try:
            ttl = self.ttl(old_key, version=version, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            error_message = f"Key '{key!r}' not found"
            raise ValueError(error_message)

        if isinstance(key, CacheKey):
            new_key = self.make_key(key.original_key(), version=version + delta)
        else:
            new_key = self.make_key(key, version=version + delta)

        return new_key, old_key, value, ttl, version

    def add(
        self: BaseClient,
        key: KeyT,
        value: EncodableT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Add a value to the cache, failing if the key already exists.

        Returns ``True`` if the object was added, ``False`` if not.
        """
        return self.set(key, value, timeout, version=version, client=client, nx=True)

    def get(
        self: BaseClient,
        key: KeyT,
        default: Any | None = None,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Any:
        """
        Retrieve a value from the cache.

        Returns decoded value if key is found, the default if not.
        """
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            value = client.get(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            return default

        return self.decode(value)

    def persist(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        try:
            return client.persist(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def expire(
        self: BaseClient,
        key: KeyT,
        timeout: ExpiryT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout  # type: ignore

        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        # for some strange reason mypy complains,
        # saying that timeout type is float | timedelta
        try:
            return client.expire(key, timeout)  # type: ignore
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def expire_at(
        self: BaseClient,
        key: KeyT,
        when: AbsExpiryT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Set an expiry flag on a ``key`` to ``when``, which can be represented
        as an integer indicating unix time or a Python datetime object.
        """
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        try:
            return client.expireat(key, when)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def pexpire(
        self: BaseClient,
        key: KeyT,
        timeout: ExpiryT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout  # type: ignore

        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        # TODO: see if the casting is necessary
        # for some strange reason mypy complains,
        # saying that timeout type is float | timedelta
        try:
            return client.pexpire(key, timeout)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def pexpire_at(
        self: BaseClient,
        key: KeyT,
        when: AbsExpiryT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Set an expiry flag on a ``key`` to ``when``, which can be represented
        as an integer indicating unix time or a Python datetime object.
        """
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        try:
            return client.pexpireat(key, when)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def get_lock(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking: bool = True,
        blocking_timeout: float | None = None,
        client: Backend | Any | None = None,
        lock_class=None,
        thread_local: bool = True,
    ) -> "Lock":
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        try:
            return client.lock(
                key,
                timeout=timeout,
                sleep=sleep,
                blocking=blocking,
                blocking_timeout=blocking_timeout,
                lock_class=lock_class,
                thread_local=thread_local,
            )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    # TODO: delete this in future releases
    lock = get_lock

    def delete(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        prefix: str | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Remove a key from the cache.
        """
        key = self.make_key(key, version=version, prefix=prefix)

        client = self._get_client(write=True, client=client, key=key)

        try:
            return client.delete(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def delete_pattern(
        self: BaseClient,
        pattern: str,
        version: int | None = None,
        prefix: str | None = None,
        client: Backend | Any | None = None,
        itersize: int | None = None,
    ) -> int:
        """
        Remove all keys matching pattern.
        """

        client = self._get_client(write=True, client=client)

        pattern = self.make_pattern(pattern, version=version, prefix=prefix)

        count = 0
        pipeline = client.pipeline()

        try:
            for key in client.scan_iter(match=pattern, count=itersize):
                pipeline.delete(key)
                count += 1
            pipeline.execute()

            return count
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def delete_many(
        self: BaseClient,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Remove multiple keys at once.
        """
        keys = [self.make_key(k, version=version) for k in keys]

        if not keys:
            return 0

        client = self._get_client(write=True, client=client)

        try:
            return client.delete(*keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def clear(self: BaseClient, client: Backend | Any | None = None) -> bool:
        """
        Flush all cache keys.
        """

        client = self._get_client(write=True, client=client)

        try:
            return client.flushdb()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def mget(
        self: BaseClient,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> dict:
        """
        atomic method.
        Retrieve many keys.
        """

        client = self._get_client(write=False, client=client)

        if not keys:
            return {}

        recovered_data = {}

        map_keys = {self.make_key(k, version=version): k for k in keys}

        try:
            results = client.mget(map_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(map_keys, results):
            if value is None:
                continue
            recovered_data[map_keys[key]] = self.decode(value)
        return recovered_data

    def get_many(
        self: BaseClient,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Backend | None = None,
    ) -> dict:
        """
        non-atomic bulk method.
        get values of the provided keys
        """
        client = self._get_client(write=False, client=client)

        pipeline = client.pipeline()
        for key in keys:
            key = self.make_key(key, version=version)
            pipeline.get(key)
        try:
            values = pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        recovered_data = {}
        for key, value in zip(keys, values):
            if not value:
                continue
            recovered_data[key] = self.decode(value)
        return recovered_data

    def set_many(
        self: BaseClient,
        data: Dict[KeyT, EncodableT],
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> None:
        """a non-atomic bulk method
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        client = self._get_client(write=True, client=client)

        pipeline = client.pipeline()
        for key, value in data.items():
            self.set(key, value, timeout, version=version, client=pipeline)
        try:
            pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def mset(
        self: BaseClient,
        data: Dict[KeyT, Any],
        timeout: float | None = None,
        version: int | None = None,
        client: Backend | None = None,
    ) -> None:
        """
        an atomic bulk method
        """
        client = self._get_client(write=True, client=client)
        data = {
            self.make_key(k, version=version): self.encode(v) for k, v in data.items()
        }
        try:
            client.mset(data)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def _incr(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
        ignore_key_check: bool = False,
    ) -> int:
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        # if key expired after exists check, then we get
        # key with wrong value and ttl -1.
        # use lua script for atomicity
        if not ignore_key_check:
            lua = """
            local exists = server.call('EXISTS', KEYS[1])
            if (exists == 1) then
                return server.call('INCRBY', KEYS[1], ARGV[1])
            else return false end
            """
        else:
            lua = """
            return server.call('INCRBY', KEYS[1], ARGV[1])
            """
        try:
            value = client.eval(lua, 1, key, delta)
            if value is None:
                error_message = f"Key '{key!r}' not found"
                raise ValueError(error_message)
        except ResponseError as e:
            # if cached value or total value is greater than 64-bit signed
            # integer.
            # elif int is encoded. so valkey sees the data as string.
            # In these situations valkey will throw ResponseError

            # try to keep TTL of key
            timeout = self.ttl(key, version=version, client=client)

            # returns -2 if the key does not exist
            # means, that key have expired
            if timeout == -2:
                error_message = f"Key '{key!r}' not found"
                raise ValueError(error_message) from e
            value = self.get(key, version=version, client=client) + delta
            self.set(key, value, version=version, timeout=timeout, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return value

    def incr(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
        ignore_key_check: bool = False,
    ) -> int:
        """
        Add delta to value in the cache. If the key does not exist, raise a
        ValueError exception. if ignore_key_check=True then the key will be
        created and set to the delta value by default.
        """
        return self._incr(
            key=key,
            delta=delta,
            version=version,
            client=client,
            ignore_key_check=ignore_key_check,
        )

    def decr(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Decrease delta to value in the cache. If the key does not exist, raise a
        ValueError exception.
        """
        return self._incr(key=key, delta=-delta, version=version, client=client)

    def ttl(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int | None:
        """
        Executes TTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            if not client.exists(key):
                return 0

            t = client.ttl(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if t >= 0:
            return t
        if t == -2:
            return 0

        # Should never reach here
        return None

    def pttl(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int | None:
        """
        Executes PTTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        key = self.make_key(key, version=version)
        client = self._get_client(write=False, client=client, key=key)

        try:
            if not client.exists(key):
                return 0

            t = client.pttl(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if t >= 0:
            return t
        if t == -2:
            return 0

        # Should never reach here
        return None

    def has_key(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Test if key exists.
        """
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)
        try:
            return client.exists(key) == 1
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def iter_keys(
        self: BaseClient,
        search: str,
        itersize: int | None = None,
        client: Backend | Any | None = None,
        version: int | None = None,
    ) -> Iterator[str]:
        """
        Same as keys, but uses cursors
        to make memory efficient keys iteration.
        """

        client = self._get_client(write=False, client=client)

        pattern = self.make_pattern(search, version=version)
        try:
            for item in client.scan_iter(match=pattern, count=itersize):
                yield self.reverse_key(item.decode())
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def keys(
        self: BaseClient,
        search: str,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> List[Any]:
        """
        Execute KEYS command and return matched results.
        Warning: this can return huge number of results, in
        this case, it strongly recommended use iter_keys
        for it.
        """

        client = self._get_client(write=False, client=client)

        pattern = self.make_pattern(search, version=version)
        try:
            return [self.reverse_key(k.decode()) for k in client.keys(pattern)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sadd(
        self: BaseClient,
        key: KeyT,
        *values: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        encoded_values = [self.encode(value) for value in values]
        try:
            return client.sadd(key, *encoded_values)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def scard(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            return client.scard(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sdiff(
        self: BaseClient,
        *keys: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        client = self._get_client(write=False, client=client)

        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return {self.decode(value) for value in client.sdiff(*nkeys)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sdiffstore(
        self: BaseClient,
        dest: KeyT,
        *keys: KeyT,
        version_dest: int | None = None,
        version_keys: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = self._get_client(write=True, client=client)

        dest = self.make_key(dest, version=version_dest)
        nkeys = [self.make_key(key, version=version_keys) for key in keys]
        try:
            return client.sdiffstore(dest, *nkeys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sinter(
        self: BaseClient,
        *keys: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        client = self._get_client(write=False, client=client)

        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return {self.decode(value) for value in client.sinter(*nkeys)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sinterstore(
        self: BaseClient,
        dest: KeyT,
        *keys: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = self._get_client(write=True, client=client)

        dest = self.make_key(dest, version=version)
        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return client.sinterstore(dest, *nkeys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def smismember(
        self: BaseClient,
        key: KeyT,
        *members: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> List[bool]:
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        encoded_members = [self.encode(member) for member in members]

        try:
            return [bool(value) for value in client.smismember(key, *encoded_members)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sismember(
        self: BaseClient,
        key: KeyT,
        member: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        member = self.encode(member)
        try:
            return bool(client.sismember(key, member))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def smembers(
        self: BaseClient,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            return {self.decode(value) for value in client.smembers(key)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def smove(
        self: BaseClient,
        source: KeyT,
        destination: KeyT,
        member: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        source = self.make_key(source, version=version)
        destination = self.make_key(destination)

        client = self._get_client(write=True, client=client, key=source)

        member = self.encode(member)
        try:
            return client.smove(source, destination, member)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def spop(
        self: BaseClient,
        key: KeyT,
        count: int | None = None,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set | Any:
        nkey = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=nkey)

        try:
            result = client.spop(nkey, count)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return self._decode_iterable_result(result)

    def srandmember(
        self: BaseClient,
        key: KeyT,
        count: int | None = None,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> List | Any:
        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            result = client.srandmember(key, count)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return self._decode_iterable_result(result, convert_to_set=False)

    def srem(
        self: BaseClient,
        key: KeyT,
        *members: EncodableT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        nmembers = [self.encode(member) for member in members]
        try:
            return client.srem(key, *nmembers)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sscan(
        self: BaseClient,
        key: KeyT,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        if self._has_compression_enabled() and match:
            err_msg = "Using match with compression is not supported."
            raise ValueError(err_msg)

        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            cursor, result = client.sscan(
                key,
                match=cast(PatternT, self.encode(match)) if match else None,
                count=count,
            )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return {self.decode(value) for value in result}

    def sscan_iter(
        self: BaseClient,
        key: KeyT,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Iterator[Any]:
        if self._has_compression_enabled() and match:
            err_msg = "Using match with compression is not supported."
            raise ValueError(err_msg)

        key = self.make_key(key, version=version)

        client = self._get_client(write=False, client=client, key=key)

        try:
            for value in client.sscan_iter(
                key,
                match=cast(PatternT, self.encode(match)) if match else None,
                count=count,
            ):
                yield self.decode(value)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sunion(
        self: BaseClient,
        *keys: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        client = self._get_client(write=False, client=client)

        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return {self.decode(value) for value in client.sunion(*nkeys)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def sunionstore(
        self: BaseClient,
        destination: Any,
        *keys: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = self._get_client(write=True, client=client)

        destination = self.make_key(destination, version=version)
        encoded_keys = [self.make_key(key, version=version) for key in keys]
        try:
            return client.sunionstore(destination, *encoded_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def close(self) -> None:
        close_flag = self._options.get(
            "CLOSE_CONNECTION",
            getattr(settings, "DJANGO_VALKEY_CLOSE_CONNECTION", False),
        )
        if close_flag:
            try:
                self._close()
            except _main_exceptions as e:
                raise ConnectionInterrupted(connection=self._clients) from e

    def _close(self) -> None:
        """
        default implementation: Override in custom client
        """
        num_clients = len(self._clients)
        for idx in range(num_clients):
            self.disconnect(index=idx)
        self._clients = [None] * num_clients

    def touch(
        self: BaseClient,
        key: KeyT,
        timeout: float | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Sets a new expiration for a key.
        """

        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        key = self.make_key(key, version=version)

        client = self._get_client(write=True, client=client, key=key)

        if timeout is None:
            try:
                return bool(client.persist(key))
            except _main_exceptions as e:
                raise ConnectionInterrupted(connection=client) from e

        # Convert to milliseconds
        timeout = int(timeout * 1000)
        try:
            return bool(client.pexpire(key, timeout))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def hset(
        self: BaseClient,
        name: str,
        key: KeyT,
        value: EncodableT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Set the value of hash name at key to value.
        Returns the number of fields added to the hash.
        """
        client = self._get_client(write=True, client=client)
        nkey = self.make_key(key, version=version)
        nvalue = self.encode(value)
        try:
            return client.hset(name, nkey, nvalue)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def hdel(
        self: BaseClient,
        name: str,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Remove keys from hash name.
        Returns the number of fields deleted from the hash.
        """
        client = self._get_client(write=True, client=client)
        nkey = self.make_key(key, version=version)
        try:
            return client.hdel(name, nkey)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def hlen(
        self: BaseClient,
        name: str,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Return the number of items in hash name.
        """
        client = self._get_client(write=False, client=client)
        try:
            return client.hlen(name)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def hkeys(
        self: BaseClient,
        name: str,
        client: Backend | Any | None = None,
    ) -> List[Any]:
        """
        Return a list of keys in hash name.
        """
        client = self._get_client(write=False, client=client)
        try:
            return [self.reverse_key(k.decode()) for k in client.hkeys(name)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def hexists(
        self: BaseClient,
        name: str,
        key: KeyT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Return True if key exists in hash name, else False.
        """
        client = self._get_client(write=False, client=client)
        nkey = self.make_key(key, version=version)
        try:
            return client.hexists(name, nkey)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e


class AsyncClientCommands(Generic[Backend]):
    def __getattr__(self, item):
        if item.startswith("a"):
            attr = getattr(self, item[1:], ATTR_DOES_NOT_EXIST)
            if attr is not ATTR_DOES_NOT_EXIST:
                return attr
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {item}"
        )

    async def _get_client(self, write=True, tried=None, client=None):
        if client:
            return client
        return await self.get_client(write, tried)

    async def get_client(
        self,
        write: bool = True,
        tried: list[int] | None = None,
    ) -> Backend | Any:
        index = self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = await self.connect(index)

        return self._clients[index]

    async def get_client_with_index(
        self, write: bool = True, tried: list[int] | None = None
    ) -> tuple[Backend, int]:
        index = self.get_next_client_index(write=write, tried=tried)

        if self._clients[index] is None:
            self._clients[index] = await self.connect(index)

        return self._clients[index], index

    async def connect(self, index: int = 0) -> Backend | Any:
        return await self.connection_factory.connect(self._server[index])

    async def disconnect(self, index: int = 0, client=None):
        if client is None:
            client = self._clients[index]

        if client is not None:
            await self.connection_factory.disconnect(client)

    async def set(
        self,
        key,
        value,
        timeout=DEFAULT_TIMEOUT,
        version=None,
        client: Backend | Any | None = None,
        nx=False,
        xx=False,
    ) -> bool:
        nkey = self.make_key(key, version=version)
        nvalue = self.encode(value)

        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        original_client = client
        tried = []

        while True:
            try:
                if client is None:
                    client, index = await self.get_client_with_index(
                        write=True, tried=tried
                    )

                if timeout is not None:
                    # convert to milliseconds
                    timeout = int(timeout) * 1000

                    if timeout <= 0:
                        if nx:
                            return not await self.has_key(
                                key, version=version, client=client
                            )

                        return bool(
                            await self.delete(key, client=client, version=version)
                        )
                return await client.set(nkey, nvalue, nx=nx, px=timeout, xx=xx)

            except _main_exceptions as e:
                if (
                    not original_client
                    and not self._replica_read_only
                    and len(tried) < len(self._server)
                ):
                    tried.append(index)
                    client = None
                    continue

                raise ConnectionInterrupted(connection=client) from e

    async def incr_version(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: Backend | None | Any = None,
    ) -> int:
        client = await self._get_client(write=True, client=client)

        new_key, old_key, value, ttl, version = await self._incr_version(
            key, delta, version, client
        )

        await self.set(new_key, value, timeout=ttl, client=client)
        await self.delete(old_key, client=client)
        return version + delta

    async def decr_version(
        self: BaseClient,
        key: KeyT,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        return await self.incr_version(
            key=key, delta=-delta, version=version, client=client
        )

    async def _incr_version(self, key, delta, version, client) -> tuple:
        if version is None:
            version = self._backend.version

        old_key = self.make_key(key, version)
        value = await self.get(old_key, version=version, client=client)

        try:
            ttl = await self.ttl(old_key, version=version, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            error_message = f"Key '{old_key!r}' does not exist"
            raise ValueError(error_message)

        if isinstance(key, CacheKey):
            new_key = self.make_key(key.original_key(), version=version + delta)
        else:
            new_key = self.make_key(key, version=version + delta)

        return new_key, old_key, value, ttl, version

    async def add(
        self,
        key,
        value,
        timeout: float | int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        return await self.set(
            key, value, timeout, version=version, client=client, nx=True
        )

    async def get(
        self,
        key,
        default: Any | None = None,
        version: int | None = None,
        client: Backend | None | Any = None,
    ) -> Any:
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)

        try:
            value = await client.get(key)

        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if value is None:
            return default

        return self.decode(value)

    async def persist(
        self, key, version: int | None = None, client: Backend | Any | None = None
    ) -> bool:
        client = await self._get_client(write=True, client=client)
        key = self.make_key(key, version=version)

        try:
            return await client.persist(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def expire(
        self,
        key,
        timeout,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        client = await self._get_client(write=True, client=client)

        key = self.make_key(key, version=version)

        try:
            return await client.expire(key, timeout)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def expire_at(
        self, key, when, version: int | None = None, client: Backend | Any | None = None
    ) -> bool:
        client = await self._get_client(write=True, client=client)

        key = self.make_key(key, version=version)

        try:
            return await client.expireat(key, when)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def pexpire(
        self,
        key,
        timeout,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        client = await self._get_client(write=True, client=client)

        key = self.make_key(key, version=version)

        try:
            return await client.pexpire(key, timeout)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def pexpire_at(
        self, key, when, version: int | None = None, client: Backend | Any | None = None
    ) -> bool:
        client = await self._get_client(write=True, client=client)

        key = self.make_key(key, version=version)

        try:
            return await client.pexpireat(key, when)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def get_lock(
        self,
        key,
        version: int | None = None,
        timeout: float | int | None = None,
        sleep: float = 0.1,
        blocking: bool = True,
        blocking_timeout: float | None = None,
        client: Backend | Any | None = None,
        lock_class=None,
        thread_local: bool = True,
    ) -> "AsyncLock":
        """Returns a Lock object, the object then should be used in an async context manager"""

        client = await self._get_client(write=True, client=client)

        key = self.make_key(key, version=version)

        try:
            return client.lock(
                key,
                timeout=timeout,
                sleep=sleep,
                blocking=blocking,
                blocking_timeout=blocking_timeout,
                lock_class=lock_class,
                thread_local=thread_local,
            )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    # TODO: delete this in future releases
    lock = aget_lock = get_lock

    async def delete(
        self,
        key,
        version: int | None = None,
        prefix: str | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = await self._get_client(write=True, client=client)

        try:
            return await client.delete(
                self.make_key(key, version=version, prefix=prefix)
            )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def delete_pattern(
        self,
        pattern: str,
        version: int | None = None,
        prefix: str | None = None,
        client: Backend | Any | None = None,
        itersize: int | None = None,
    ) -> int:
        """
        Remove all keys matching a pattern.
        """
        client = await self._get_client(write=True, client=client)

        pattern = self.make_pattern(pattern, version=version, prefix=prefix)

        count = 0
        pipeline = await client.pipeline()

        try:
            async with contextlib.aclosing(
                client.scan_iter(match=pattern, count=itersize)
            ) as values:
                async for key in values:
                    await pipeline.delete(key)
                    count += 1
                await pipeline.execute()

            return count

        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def delete_many(
        self, keys, version: int | None = None, client: Backend | None = None
    ) -> int:
        """Remove multiple keys at once."""
        keys = [self.make_key(k, version=version) for k in keys]

        if not keys:
            return 0

        client = await self._get_client(write=True, client=client)

        try:
            return await client.delete(*keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def clear(self, client: Backend | Any | None = None) -> bool:
        """
        Flush all cache keys.
        """

        client = await self._get_client(write=True, client=client)

        try:
            return await client.flushdb()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def mget(
        self, keys, version: int | None = None, client: Backend | Any | None = None
    ) -> dict:
        if not keys:
            return {}

        client = await self._get_client(write=False, client=client)

        recovered_data = {}

        map_keys = {self.make_key(k, version=version): k for k in keys}

        try:
            results = await client.mget(*map_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(map_keys, results):
            if value is None:
                continue
            recovered_data[map_keys[key]] = self.decode(value)

        return recovered_data

    async def get_many(
        self,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Backend | None = None,
    ):
        """
        non-atomic bulk method.
        get values of the provided keys.
        """
        client = await self._get_client(write=False, client=client)

        pipeline = await client.pipeline()
        for key in keys:
            key = self.make_key(key, version=version)
            await pipeline.get(key)
        try:
            values = await pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        recovered_data = {}
        for key, value in zip(keys, values):
            if not value:
                continue
            recovered_data[key] = self.decode(value)
        return recovered_data

    async def set_many(
        self,
        data: dict,
        timeout: float | int | None = None,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> None:
        client = await self._get_client(write=True, client=client)

        pipeline = await client.pipeline()
        for key, value in data.items():
            await self.set(key, value, timeout, version=version, client=pipeline)
        try:
            await pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def mset(
        self,
        data: Dict[KeyT, Any],
        timeout: float | None = None,
        version: int | None = None,
        client: Backend | None = None,
    ) -> None:
        client = await self._get_client(write=True, client=client)

        data = {
            self.make_key(k, version=version): self.encode(v) for k, v in data.items()
        }

        try:
            await client.mset(data)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def _incr(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
        ignoree_key_check: bool = False,
    ) -> int:
        client = await self._get_client(write=True, client=client)
        key = self.make_key(key, version=version)

        # if key expired after exists check, then we get
        # key with wrong value and ttl -1.
        # use lua script for atomicity
        if not ignoree_key_check:
            lua = """
            local exists = server.call('EXISTS', KEYS[1])
            if (exists == 1) then
                return server.call('INCRBY', KEYS[1], ARGV[1])
            else return false end
            """
        else:
            lua = """
            return server.call('INCRBY', KEYS[1], ARGV[1])
            """
        try:
            value = await client.eval(lua, 1, key, delta)
            if value is None:
                error_message = f"Key '{key!r}' not found"
                raise ValueError(error_message)
        except ResponseError as e:
            # if cached value or total value is greater than 64-bit signed
            # integer.
            # elif int is encoded. so valkey sees the data as string.
            # In these situations valkey will throw ResponseError

            # try to keep TTL of key
            timeout = await self.ttl(key, version=version, client=client)

            if timeout == -2:
                error_message = f"Key '{key!r}' not found"
                raise ValueError(error_message) from e
            value = await self.get(key, version=version, client=client) + delta
            await self.set(key, value, version=version, timeout=timeout, client=client)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        return value

    async def incr(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
        ignore_key_check: bool = False,
    ) -> int:
        """
        Add delta to value in the cache. If the key does not exist, raise a
        ValueError exception. if ignore_key_check=True then the key will be
        created and set to the delta value by default.
        """
        return await self._incr(
            key,
            delta,
            version=version,
            client=client,
            ignoree_key_check=ignore_key_check,
        )

    async def decr(
        self,
        key,
        delta: int = 1,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Decrease delta to value in the cache. If the key does not exist, raise a
        ValueError exception.
        """
        return await self._incr(key, delta=-delta, version=version, client=client)

    async def ttl(
        self, key, version: int | None = None, client: Backend | Any | None = None
    ) -> int | None:
        """
        Executes TTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        client = await self._get_client(write=False, client=client)
        key = self.make_key(key, version=version)
        try:
            if not await client.exists(key):
                return 0

            t = await client.ttl(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        if t >= 0:
            return t
        if t == -2:
            return 0

        return None

    async def pttl(
        self, key, version: int | None = None, client: Backend | Any | None = None
    ) -> int | None:
        """
        Executes PTTL valkey command and return the "time-to-live" of specified key.
        If key is a non-volatile key, it returns None.
        """
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        try:
            if not await client.exists(key):
                return 0

            t = await client.pttl(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        if t >= 0:
            return t
        if t == -2:
            return 0

        return None

    async def has_key(
        self, key, version: int | None = None, client: Backend | Any | None = None
    ) -> bool:
        """
        Test if key exists.
        """
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        try:
            return await client.exists(key) == 1
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def iter_keys(
        self,
        search: str,
        itersize: int | None = None,
        client: Backend | Any | None = None,
        version: int | None = None,
    ) -> AsyncGenerator[str, None]:
        """
        Same as keys, but uses cursors
        to make memory efficient keys iteration.
        """
        client = await self._get_client(write=False, client=client)
        pattern = self.make_pattern(search, version=version)
        try:
            async with contextlib.aclosing(
                client.scan_iter(match=pattern, count=itersize)
            ) as values:
                async for item in values:
                    yield self.reverse_key(item.decode())
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def keys(
        self,
        search: str,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> list[Any]:
        client = await self._get_client(write=False, client=client)

        pattern = self.make_pattern(search, version=version)
        try:
            return [self.reverse_key(k.decode()) for k in await client.keys(pattern)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sadd(
        self,
        key,
        *values: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = await self._get_client(write=True, client=client)
        key = self.make_key(key, version=version)
        encoded_values = [self.encode(value) for value in values]

        try:
            return await client.sadd(key, *encoded_values)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def scard(
        self, key, version: int | None = None, client: Backend | Any | None = None
    ) -> int:
        client = await self._get_client(write=False, client=client)
        key = self.make_key(key, version=version)
        try:
            return await client.scard(key)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sdiff(
        self, *keys, version: int | None = None, client: Backend | Any | None = None
    ) -> Set[Any]:
        client = await self._get_client(write=False, client=client)
        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return {self.decode(value) for value in await client.sdiff(*nkeys)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sdiffstore(
        self,
        dest,
        *keys,
        version_dest: int | None = None,
        version_keys: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = await self._get_client(write=True, client=client)
        dest = self.make_key(dest, version=version_dest)
        nkeys = [self.make_key(key, version=version_keys) for key in keys]
        try:
            return await client.sdiffstore(dest, *nkeys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sinter(
        self, *keys, version: int | None = None, client: Backend | Any | None = None
    ) -> Set[Any]:
        client = await self._get_client(write=False, client=client)
        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return {self.decode(value) for value in await client.sinter(*nkeys)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sinterstore(
        self,
        dest,
        *keys,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = await self._get_client(write=True, client=client)
        dest = self.make_key(dest, version=version)
        nkeys = [self.make_key(key, version=version) for key in keys]

        try:
            return await client.sinterstore(dest, *nkeys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def smismember(
        self,
        key,
        *members: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> list[bool]:
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        encoded_members = [self.encode(member) for member in members]

        try:
            return [
                bool(value) for value in await client.smismember(key, *encoded_members)
            ]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sismember(
        self,
        key,
        member: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        member = self.encode(member)
        try:
            return bool(await client.sismember(key, member))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def smembers(
        self, key, version: int | None = None, client: Backend | Any | None = None
    ) -> Set[Any]:
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        try:
            return {self.decode(value) for value in await client.smembers(key)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def smove(
        self,
        source,
        destination,
        member: Any,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        client = await self._get_client(write=False, client=client)
        source = self.make_key(source, version=version)
        destination = self.make_key(destination, version=version)
        member = self.encode(member)
        try:
            return await client.smove(source, destination, member)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def spop(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set | Any:
        client = await self._get_client(write=True, client=client)
        nkey = self.make_key(key, version=version)
        try:
            result = await client.spop(nkey, count)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return self._decode_iterable_result(result)

    async def srandmember(
        self,
        key,
        count: int | None = None,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> list | Any:
        client = await self._get_client(write=False, client=client)
        key = self.make_key(key, version=version)
        try:
            result = await client.srandmember(key, count)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return self._decode_iterable_result(result, convert_to_set=False)

    async def srem(
        self,
        key,
        *members,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        nmembers = [self.encode(member) for member in members]
        try:
            return await client.srem(key, *nmembers)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sscan(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        # TODO check this is correct
        if self._has_compression_enabled() and match:
            error_message = "Using match with compression is not supported."
            raise ValueError(error_message)

        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)
        try:
            cursor, result = await client.sscan(
                key,
                match=cast(PatternT, self.encode(match)) if match else None,
                count=count,
            )
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e
        return {self.decode(value) for value in result}

    async def sscan_iter(
        self,
        key,
        match: str | None = None,
        count: int = 10,
        version: int | None = None,
        client: Backend | Any | None = None,
    ):
        if self._has_compression_enabled() and match:
            error_message = "Using match with compression is not supported."
            raise ValueError(error_message)

        client = await self._get_client(write=False, client=client)

        key = self.make_key(key, version=version)

        try:
            async with contextlib.aclosing(
                client.sscan_iter(
                    key,
                    match=cast(PatternT, self.encode(match)) if match else None,
                    count=count,
                )
            ) as values:
                async for value in values:
                    yield self.decode(value)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sunion(
        self,
        *keys,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> Set[Any]:
        client = await self._get_client(write=False, client=client)

        nkeys = [self.make_key(key, version=version) for key in keys]
        try:
            return {self.decode(value) for value in await client.sunion(*nkeys)}
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def sunionstore(
        self,
        destination: Any,
        *keys,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        client = await self._get_client(write=True, client=client)
        destination = self.make_key(destination, version=version)
        encoded_keys = [self.make_key(key, version=version) for key in keys]
        try:
            return await client.sunionstore(destination, *encoded_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def close(self) -> None:
        close_flag = self._options.get(
            "CLOSE_CONNECTION",
            getattr(settings, "DJANGO_VALKEY_CLOSE_CONNECTION", False),
        )
        if close_flag:
            try:
                await self._close()
            except _main_exceptions as e:
                raise ConnectionInterrupted(connection=self._clients) from e

    async def _close(self) -> None:
        """
        default implementation: Override in custom client
        """
        num_clients = len(self._clients)
        for index in range(num_clients):
            # TODO: check disconnect and close
            await self.disconnect(index=index)
        self._clients = [None] * num_clients

    async def touch(
        self,
        key,
        timeout: float | int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Sets a new expiration for a key.
        """
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        client = await self._get_client(write=True, client=client)

        key = self.make_key(key, version=version)
        if timeout is None:
            try:
                return bool(await client.persist(key))
            except _main_exceptions as e:
                raise ConnectionInterrupted(connection=client) from e

        # convert timeout to milliseconds
        timeout = int(timeout * 1000)
        try:
            return bool(await client.pexpire(key, timeout))
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def hset(
        self,
        name: str,
        key,
        value,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Sets the value of hash name at key to value.
        Returns the number of fields added to the hash.
        """
        client = await self._get_client(write=True, client=client)

        nkey = self.make_key(key, version)
        nvalue = self.encode(value)

        try:
            return await client.hset(name, nkey, nvalue)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def hdel(
        self,
        name: str,
        key,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> int:
        """
        Remove keys from hash name.
        Returns the number of fields deleted from the hash.
        """
        client = await self._get_client(write=True, client=client)
        nkey = self.make_key(key, version=version)
        try:
            return await client.hdel(name, nkey)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def hlen(self, name: str, client: Backend | Any | None = None) -> int:
        """
        Return the number of items in hash name.
        """
        client = await self._get_client(write=False, client=client)
        try:
            return await client.hlen(name)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def hkeys(self, name: str, client: Backend | Any | None = None) -> list[Any]:
        client = await self._get_client(write=False, client=client)

        try:
            return [self.reverse_key(k.decode()) for k in await client.hkeys(name)]
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    async def hexists(
        self,
        name: str,
        key,
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> bool:
        """
        Return True if key exists in hash name, else False.
        """
        client = await self._get_client(write=False, client=client)
        nkey = self.make_key(key, version=version)
        try:
            return await client.hexists(name, nkey)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e


# Herd related code:
class Marker:
    """
    Dummy class for use as
    marker for herded keys.
    """

    pass


def _is_expired(x, herd_timeout: int) -> bool:
    if x >= herd_timeout:
        return True
    val = x + random.randint(1, herd_timeout)

    if val >= herd_timeout:
        return True
    return False


class HerdCommonMethods:
    def __init__(self, *args, **kwargs):
        self._marker = Marker()
        self._herd_timeout: int = getattr(settings, "CACHE_HERD_TIMEOUT", 60)
        super().__init__(*args, **kwargs)

    def _pack(self, value: Any, timeout) -> tuple[Marker, Any, int]:
        herd_timeout = (timeout or self._backend.default_timeout) + int(time.time())
        return self._marker, value, herd_timeout

    def _unpack(self, value: tuple[Marker, Any, int]) -> tuple[Any, bool]:
        try:
            marker, unpacked, herd_timeout = value
        except (ValueError, TypeError):
            return value, False

        if not isinstance(marker, Marker):
            return value, False

        now = time.time()
        if herd_timeout < now:
            x = now - herd_timeout
            return unpacked, _is_expired(x, self._herd_timeout)

        return unpacked, False



from valkey import Valkey

from django_valkey.base_client import BaseClient, ClientCommands


class DefaultClient(BaseClient[Valkey], ClientCommands[Valkey]):
    CONNECTION_FACTORY_PATH = "django_valkey.pool.ConnectionFactory"



from collections.abc import Iterable

import pytest

from django_valkey.cluster_cache.cache import ClusterValkeyCache
from django_valkey.cluster_cache.client import DefaultClusterClient


@pytest.fixture
def cache_client(cache: ClusterValkeyCache) -> Iterable[DefaultClusterClient]:
    client = cache.client
    client.set("TestClientClose", 0)
    yield client
    client.delete("TestClientClose")


class TestDefaultClusterClient:
    def test_delete_pattern_calls_get_client_given_no_client(self, mocker):
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client"
        )
        client = DefaultClusterClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""

        client.delete_pattern(pattern="{foo}*")
        get_client_mock.assert_called_once_with(write=True, tried=None)

    def test_delete_pattern_calls_make_pattern(self, mocker):
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client",
            return_value=mocker.Mock(),
        )
        make_pattern_mock = mocker.patch(
            "django_valkey.base_client.BaseClient.make_pattern"
        )
        client = DefaultClusterClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        get_client_mock.return_value.scan_iter.return_value = []

        client.delete_pattern(pattern="{foo}*")

        kwargs = {"version": None, "prefix": None}
        make_pattern_mock.assert_called_once_with("{foo}*", **kwargs)

    def test_delete_pattern_calls_scan_iter_with_count_if_itersize_given(self, mocker):
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client",
            return_value=mocker.Mock(),
        )
        make_pattern_mock = mocker.patch(
            "django_valkey.base_client.BaseClient.make_pattern"
        )
        client = DefaultClusterClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        get_client_mock.return_value.scan_iter.return_value = []

        client.delete_pattern(pattern="{foo}*", itersize=90210)

        get_client_mock.return_value.scan_iter.assert_called_once_with(
            count=90210, match=make_pattern_mock.return_value
        )

    def test_delete_pattern_calls_pipeline_delete_and_execute(self, mocker):
        mocker.patch("django_valkey.base_client.BaseClient.__init__", return_value=None)
        get_client_mock = mocker.patch(
            "django_valkey.base_client.ClientCommands.get_client",
            return_value=mocker.Mock(),
        )
        mocker.patch("django_valkey.base_client.BaseClient.make_pattern")

        client = DefaultClusterClient()
        client._backend = mocker.Mock()
        client._backend.key_prefix = ""
        get_client_mock.return_value.scan_iter.return_value = [":1:{foo}", ":1:{foo}-a"]
        get_client_mock.return_value.pipeline.return_value = mocker.Mock()
        get_client_mock.return_value.pipeline.return_value.delete = mocker.Mock()
        get_client_mock.return_value.pipeline.return_value.execute = mocker.Mock()

        client.delete_pattern(pattern="{foo}*")

        assert get_client_mock.return_value.pipeline.return_value.delete.call_count == 2
        get_client_mock.return_value.pipeline.return_value.delete.assert_has_calls(
            [mocker.call(":1:{foo}"), mocker.call(":1:{foo}-a")]
        )
        get_client_mock.return_value.pipeline.return_value.execute.assert_called_once()



from collections.abc import Iterable

import pytest
from pytest_django.fixtures import SettingsWrapper

from django.core.cache import caches
from django.test import override_settings

from valkey.cluster import ValkeyCluster

from django_valkey.cluster_cache.cache import ClusterValkeyCache


@pytest.fixture
def patch_itersize_setting() -> Iterable[None]:
    # destroy cache to force recreation with overridden settings
    del caches["default"]
    with override_settings(DJANGO_VALKEY_SCAN_ITERSIZE=30):
        yield
    # destroy cache to force recreation with original settings
    del caches["default"]


class TestDjangoValkeyCache:
    def test_mget(self, cache: ClusterValkeyCache):
        cache.set("{foo}a", 1)
        cache.set("{foo}b", 2)
        cache.set("{foo}c", 3)

        res = cache.mget(["{foo}a", "{foo}b", "{foo}c"])
        assert res == {"{foo}a": 1, "{foo}b": 2, "{foo}c": 3}

    def test_mget_unicode(self, cache: ClusterValkeyCache):
        cache.set("{foo}a", "1")
        cache.set("{foo}ب", "2")
        cache.set("{foo}c", "الف")

        res = cache.mget(["{foo}a", "{foo}ب", "{foo}c"])
        assert res == {"{foo}a": "1", "{foo}ب": "2", "{foo}c": "الف"}

    def test_mset(self, cache: ClusterValkeyCache):
        cache.mset({"a{foo}": 1, "b{foo}": 2, "c{foo}": 3})
        res = cache.mget(["a{foo}", "b{foo}", "c{foo}"])
        assert res == {"a{foo}": 1, "b{foo}": 2, "c{foo}": 3}

    def test_msetnx(self, cache: ClusterValkeyCache):
        cache.mset({"a{foo}": 1, "b{foo}": 2, "c{foo}": 3})
        res = cache.mget(["a{foo}", "b{foo}", "c{foo}"])
        assert res == {"a{foo}": 1, "b{foo}": 2, "c{foo}": 3}
        cache.msetnx({"a{foo}": 3, "new{foo}": 1, "other{foo}": 1})
        res = cache.mget(["a{foo}", "new{foo}", "other{foo}"])
        assert res == {"a{foo}": 1}

    def test_delete_pattern(self, cache: ClusterValkeyCache):
        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            cache.set(key, "foo")

        res = cache.delete_pattern("*foo-a*")
        assert bool(res) is True

        keys = cache.keys("foo*", target_nodes=ValkeyCluster.ALL_NODES)
        assert set(keys) == {"foo-bb", "foo-bc"}

        res = cache.delete_pattern("*foo-a*")
        assert bool(res) is False

    def test_delete_pattern_with_custom_count(self, cache: ClusterValkeyCache, mocker):
        client_mock = mocker.patch(
            "django_valkey.cluster_cache.cache.ClusterValkeyCache.client"
        )
        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            cache.set(key, "foo")

        cache.delete_pattern("*foo-a*", itersize=2)

        client_mock.delete_pattern.assert_called_once_with("*foo-a*", itersize=2)

    def test_delete_pattern_with_settings_default_scan_count(
        self,
        patch_itersize_setting,
        cache: ClusterValkeyCache,
        settings: SettingsWrapper,
        mocker,
    ):
        client_mock = mocker.patch(
            "django_valkey.cluster_cache.cache.ClusterValkeyCache.client"
        )
        for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
            cache.set(key, "foo")
        expected_count = settings.DJANGO_VALKEY_SCAN_ITERSIZE

        cache.delete_pattern("*foo-a*")

        client_mock.delete_pattern.assert_called_once_with(
            "*foo-a*", itersize=expected_count
        )

    def test_sdiff(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.sdiff("{foo}1", "{foo}2") == {"bar1"}

    def test_sdiffstore(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.sdiffstore("{foo}3", "{foo}1", "{foo}2") == 1
        assert cache.smembers("{foo}3") == {"bar1"}

    def test_sdiffstore_with_keys_version(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2", version=2)
        cache.sadd("{foo}2", "bar2", "bar3", version=2)
        assert cache.sdiffstore("{foo}3", "{foo}1", "{foo}2", version_keys=2) == 1
        assert cache.smembers("{foo}3") == {"bar1"}

    def test_sdiffstore_with_different_keys_versions_without_initial_set_in_version(
        self, cache: ClusterValkeyCache
    ):
        cache.sadd("{foo}1", "bar1", "bar2", version=1)
        cache.sadd("{foo}2", "bar2", "bar3", version=2)
        assert cache.sdiffstore("{foo}3", "{foo}1", "{foo}2", version_keys=2) == 0

    def test_sdiffstore_with_different_keys_versions_with_initial_set_in_version(
        self, cache: ClusterValkeyCache
    ):
        cache.sadd("{foo}1", "bar1", "bar2", version=2)
        cache.sadd("{foo}2", "bar2", "bar3", version=1)
        assert cache.sdiffstore("{foo}3", "{foo}1", "{foo}2", version_keys=2) == 2

    def test_sinter(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.sinter("{foo}1", "{foo}2") == {"bar2"}

    def test_sinterstore(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.sinterstore("{foo}3", "{foo}1", "{foo}2") == 1
        assert cache.smembers("{foo}3") == {"bar2"}

    def test_smove(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.smove("{foo}1", "{foo}2", "bar1") is True
        assert cache.smove("{foo}1", "{foo}2", "bar4") is False
        assert cache.smembers("{foo}1") == {"bar2"}
        assert cache.smembers("{foo}2") == {"bar1", "bar2", "bar3"}

    def test_sunion(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.sunion("{foo}1", "{foo}2") == {"bar1", "bar2", "bar3"}

    def test_sunionstore(self, cache: ClusterValkeyCache):
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.sadd("{foo}2", "bar2", "bar3")
        assert cache.sunionstore("{foo}3", "{foo}1", "{foo}2") == 3
        assert cache.smembers("{foo}3") == {"bar1", "bar2", "bar3"}

    def test_flushall(self, cache: ClusterValkeyCache):
        cache.set("{foo}a", 1)
        cache.sadd("{foo}1", "bar1", "bar2")
        cache.hset("foo_hash1", "foo1", "bar1")
        cache.flushall()
        assert not cache.get("{foo}a")
        assert cache.smembers("{foo}a") == set()
        assert not cache.hexists("foo_hash1", "foo1")



import builtins
import contextlib
import functools
import inspect
import logging
from collections.abc import AsyncGenerator, Callable, Iterator
from typing import (
    Any,
    TypeVar,
    Generic,
    TYPE_CHECKING,
)

from django.conf import settings
from django.core.cache.backends.base import get_key_func
from django.utils.module_loading import import_string

from django_valkey.exceptions import ConnectionInterrupted

if TYPE_CHECKING:
    from valkey.lock import Lock
    from valkey.asyncio.lock import Lock as ALock

Client = TypeVar("Client")
Backend = TypeVar("Backend")

CONNECTION_INTERRUPTED = object()
DEFAULT_TIMEOUT = object()
ATTR_DOES_NOT_EXIST = object()


def decorate_all_methods(decorator):
    def decorate(cls):
        for attr in dir(cls):
            # dunders and `get` should not be decorated
            # get is handled by `_get`
            if attr.startswith("__") or attr in {"get", "get_or_set"}:
                continue

            attribute = getattr(cls, attr)
            if callable(attribute):
                if attr == "_get":
                    setattr(
                        cls,
                        attr,
                        decorator(attribute, return_value=CONNECTION_INTERRUPTED),
                    )
                elif attr == "get_many":
                    setattr(cls, attr, decorator(attribute, return_value={}))
                elif attr in {"iter_keys", "sscan_iter"}:
                    setattr(cls, attr, decorator(attribute, gen=True))
                else:
                    setattr(cls, attr, decorator(attribute))
        return cls

    return decorate


def omit_exception(
    method: Callable | None = None, return_value: Any | None = None, gen: bool = False
):
    """
    Simple decorator that intercepts connection
    errors and ignores these if settings specify this.
    """

    if method is None:
        return functools.partial(omit_exception, return_value=return_value)

    def __handle_error(self, e) -> Any | None:
        if getattr(self, "_ignore_exceptions", None):
            if getattr(self, "_log_ignored_exceptions", None):
                self.logger.exception("Exception ignored")

            return return_value
        raise e.__cause__

    @functools.wraps(method)
    def _decorator(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            return __handle_error(self, e)

    @functools.wraps(method)
    def _generator_decorator(self, *args, **kwargs):
        try:
            for item in method(self, *args, **kwargs):
                yield item
        except ConnectionInterrupted as e:
            yield __handle_error(self, e)

    # sync generators (iter_keys, sscan_iter) are only generator on client class
    # in the backend they are just a function (that returns a generator)
    # so inspect.isgeneratorfunction does not work
    if not gen:
        wrapper = _decorator

    # if method is a generator or async generator, it should be iterated over by this decorator
    # generators don't error by simply being called, they need to be iterated over.
    else:
        wrapper = _generator_decorator

    return wrapper


def omit_exception_async(
    method: Callable | None = None, return_value: Any | None = None, gen=False
):
    if method is None:
        return functools.partial(omit_exception, return_value=return_value)

    def __handle_error(self, e) -> Any | None:
        if getattr(self, "_ignore_exceptions", None):
            if getattr(self, "_log_ignored_exceptions", None):
                self.logger.exception("Exception ignored")

            return return_value
        raise e.__cause__

    @functools.wraps(method)
    async def _async_decorator(self, *args, **kwargs):
        try:
            return await method(self, *args, **kwargs)
        except ConnectionInterrupted as e:
            return __handle_error(self, e)

    @functools.wraps(method)
    async def _async_generator_decorator(self, *args, **kwargs):
        try:
            async for item in method(self, *args, **kwargs):
                yield item
        except ConnectionInterrupted as e:
            yield __handle_error(self, e)

    if inspect.isasyncgenfunction(method):
        # if method is a generator or async generator, it should be iterated over by this decorator
        # generators don't error by simply being called, they need to be iterated over.
        wrapper = _async_generator_decorator
    else:
        wrapper = _async_decorator

    return wrapper


class BaseValkeyCache(Generic[Client, Backend]):
    _missing_key = object()

    def __init__(self, server: str, params: dict[str, Any]) -> None:
        timeout = params.get("timeout", params.get("TIMEOUT", 300))
        if timeout is not None:
            try:
                timeout = int(timeout)
            except (ValueError, TypeError):
                timeout = 300
        self.default_timeout = timeout

        options = params.get("OPTIONS", {})
        max_entries = params.get("max_entries", options.get("MAX_ENTRIES", 300))
        try:
            self._max_entries = int(max_entries)
        except (ValueError, TypeError):
            self._max_entries = 300

        cull_frequency = params.get("cull_frequency", options.get("CULL_FREQUENCY", 3))
        try:
            self._cull_frequency = int(cull_frequency)
        except (ValueError, TypeError):
            self._cull_frequency = 3

        self.key_prefix = params.get("KEY_PREFIX", "")
        self.version = params.get("VERSION", 1)
        self.key_func = get_key_func(params.get("KEY_FUNCTION"))

        self._server = server
        self._params = params
        self._default_scan_itersize = getattr(
            settings, "DJANGO_VALKEY_SCAN_ITERSIZE", 10
        )

        options: dict = params.get("OPTIONS", {})
        self._client_cls = self.get_client_class()
        self._client = None

        self._ignore_exceptions = options.get(
            "IGNORE_EXCEPTIONS",
            getattr(settings, "DJANGO_VALKEY_IGNORE_EXCEPTIONS", False),
        )
        self._log_ignored_exceptions = options.get(
            "LOG_IGNORE_EXCEPTIONS",
            getattr(settings, "DJANGO_VALKEY_LOG_IGNORED_EXCEPTIONS", False),
        )
        self.logger = (
            logging.getLogger(getattr(settings, "DJANGO_VALKEY_LOGGER", __name__))
            if self._log_ignored_exceptions
            else None
        )

    def get_client_class(self) -> type[Client] | type:
        options = self._params.get("OPTIONS", {})
        _client_cls = options.get("CLIENT_CLASS", self.DEFAULT_CLIENT_CLASS)
        return import_string(_client_cls)

    @property
    def client(self) -> Client:
        """
        Lazy client connection property.
        """
        if self._client is None:
            self._client = self._client_cls(self._server, self._params, self)
        return self._client

    def make_key(self, *args, **kwargs) -> bool:
        return self.client.make_key(*args, **kwargs)

    def make_pattern(self, *args, **kwargs) -> bool:
        return self.client.make_pattern(*args, **kwargs)


class BackendCommands:
    def __contains__(self, item):
        return self.has_key(item)

    def set(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.set(*args, **kwargs)

    def incr_version(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.incr_version(*args, **kwargs)

    def decr_version(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.decr_version(*args, **kwargs)

    def add(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.add(*args, **kwargs)

    def get(self, key, default=None, version=None, client=None) -> Any:
        value = self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default
        return value

    def _get(self: BaseValkeyCache, key, default, version, client) -> Any:
        return self.client.get(key, default=default, version=version, client=client)

    def get_or_set(self, key, default, timeout=DEFAULT_TIMEOUT, version=None) -> Any:
        """
        Fetch a given key from the cache. If the key does not exist,
        add the key and set it to the default value. The default value can
        also be any callable. If timeout is given, use that timeout for the
        key; otherwise use the default cache timeout.

        Return the value of the key stored or retrieved.
        """
        val = self.get(key, self._missing_key, version=version)
        if val is self._missing_key:
            if callable(default):
                default = default()
            self.add(key, default, timeout=timeout, version=version)
            # Fetch the value again to avoid a race condition if another caller
            # added a value between the first get() and the add() above.
            return self.get(key, default, version=version)
        return val

    def delete(self: BaseValkeyCache, *args, **kwargs) -> int:
        result = self.client.delete(*args, **kwargs)
        return bool(result)

    def delete_many(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.delete_many(*args, **kwargs)

    def delete_pattern(self: BaseValkeyCache, *args, **kwargs) -> int:
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return self.client.delete_pattern(*args, **kwargs)

    def clear(self: BaseValkeyCache) -> bool:
        return self.client.clear()

    def get_many(self: BaseValkeyCache, *args, **kwargs) -> dict:
        return self.client.get_many(*args, **kwargs)

    def set_many(self: BaseValkeyCache, *args, **kwargs) -> None:
        return self.client.set_many(*args, **kwargs)

    def incr(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.incr(*args, **kwargs)

    def decr(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.decr(*args, **kwargs)

    def has_key(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.has_key(*args, **kwargs)

    def keys(self: BaseValkeyCache, *args, **kwargs) -> list[Any]:
        return self.client.keys(*args, **kwargs)

    def iter_keys(self: BaseValkeyCache, *args, **kwargs) -> Iterator[Any]:
        return self.client.iter_keys(*args, **kwargs)

    def close(self: BaseValkeyCache) -> None:
        self.client.close()

    def touch(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.touch(*args, **kwargs)

    def mset(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.mset(*args, **kwargs)

    def mget(self: BaseValkeyCache, *args, **kwargs) -> dict | list[Any]:
        return self.client.mget(*args, **kwargs)

    def persist(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.persist(*args, **kwargs)

    def expire(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.expire(*args, **kwargs)

    def expire_at(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.expire_at(*args, **kwargs)

    def pexpire(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.pexpire(*args, **kwargs)

    def pexpire_at(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.pexpire_at(*args, **kwargs)

    def get_lock(self: BaseValkeyCache, *args, **kwargs) -> "Lock":
        return self.client.get_lock(*args, **kwargs)

    lock = get_lock

    def ttl(self: BaseValkeyCache, *args, **kwargs) -> int | None:
        return self.client.ttl(*args, **kwargs)

    def pttl(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.pttl(*args, **kwargs)

    def sadd(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.sadd(*args, **kwargs)

    def scard(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.scard(*args, **kwargs)

    def sdiff(self: BaseValkeyCache, *args, **kwargs) -> builtins.set[Any]:
        return self.client.sdiff(*args, **kwargs)

    def sdiffstore(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.sdiffstore(*args, **kwargs)

    def sinter(self: BaseValkeyCache, *args, **kwargs) -> builtins.set[Any]:
        return self.client.sinter(*args, **kwargs)

    def sinterstore(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.sinterstore(*args, **kwargs)

    def smismember(self: BaseValkeyCache, *args, **kwargs) -> list[bool]:
        return self.client.smismember(*args, **kwargs)

    def sismember(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.sismember(*args, **kwargs)

    def smembers(self: BaseValkeyCache, *args, **kwargs) -> builtins.set[Any]:
        return self.client.smembers(*args, **kwargs)

    def smove(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.smove(*args, **kwargs)

    def spop(self: BaseValkeyCache, *args, **kwargs) -> builtins.set | Any:
        return self.client.spop(*args, **kwargs)

    def srandmember(self: BaseValkeyCache, *args, **kwargs) -> list | Any:
        return self.client.srandmember(*args, **kwargs)

    def srem(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.srem(*args, **kwargs)

    def sscan(self: BaseValkeyCache, *args, **kwargs) -> builtins.set[Any]:
        return self.client.sscan(*args, **kwargs)

    def sscan_iter(self: BaseValkeyCache, *args, **kwargs) -> Iterator:
        return self.client.sscan_iter(*args, **kwargs)

    def sunion(self: BaseValkeyCache, *args, **kwargs) -> builtins.set[Any]:
        return self.client.sunion(*args, **kwargs)

    def sunionstore(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.sunionstore(*args, **kwargs)

    def hset(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.hset(*args, **kwargs)

    def hdel(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.hdel(*args, **kwargs)

    def hlen(self: BaseValkeyCache, *args, **kwargs) -> int:
        return self.client.hlen(*args, **kwargs)

    def hkeys(self: BaseValkeyCache, *args, **kwargs) -> list[Any]:
        return self.client.hkeys(*args, **kwargs)

    def hexists(self: BaseValkeyCache, *args, **kwargs) -> bool:
        return self.client.hexists(*args, **kwargs)


class AsyncBackendCommands:
    def __getattr__(self, item):
        if item.startswith("a"):
            attr = getattr(self, item[1:], ATTR_DOES_NOT_EXIST)
            if attr is not ATTR_DOES_NOT_EXIST:
                return attr
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {item}"
        )

    async def set(self: BaseValkeyCache, *args, **kwargs):
        return await self.client.aset(*args, **kwargs)

    async def incr_version(self, *args, **kwargs):
        return await self.client.aincr_version(*args, **kwargs)

    async def decr_version(self, *args, **kwargs):
        return await self.client.adecr_version(*args, **kwargs)

    async def add(self, *args, **kwargs):
        return await self.client.aadd(*args, **kwargs)

    async def get(self, key, default=None, version=None, client=None):
        value = await self._get(key, default, version, client)
        if value is CONNECTION_INTERRUPTED:
            value = default
        return value

    async def _get(self, key, default=None, version=None, client=None):
        return await self.client.aget(key, default, version, client)

    async def get_or_set(self, key, default, timeout=DEFAULT_TIMEOUT, version=None):
        """See get_or_set()."""
        val = await self.aget(key, self._missing_key, version=version)
        if val is self._missing_key:
            if callable(default):
                default = default()
            await self.aadd(key, default, timeout=timeout, version=version)
            # Fetch the value again to avoid a race condition if another caller
            # added a value between the first aget() and the aadd() above.
            return await self.aget(key, default, version=version)
        return val

    async def delete(self, *args, **kwargs):
        result = await self.client.adelete(*args, **kwargs)
        return bool(result)

    async def delete_many(self, *args, **kwargs):
        return await self.client.adelete_many(*args, **kwargs)

    async def clear(self):
        return await self.client.aclear()

    async def get_many(self, *args, **kwargs):
        return await self.client.aget_many(*args, **kwargs)

    async def set_many(self, *args, **kwargs):
        return await self.client.aset_many(*args, **kwargs)

    async def incr(self, *args, **kwargs):
        return await self.client.aincr(*args, **kwargs)

    async def decr(self, *args, **kwargs):
        return await self.client.adecr(*args, **kwargs)

    async def has_key(self, *args, **kwargs):
        return await self.client.ahas_key(*args, **kwargs)

    async def close(self, *args, **kwargs):
        return await self.client.close()

    async def touch(self, *args, **kwargs):
        return await self.client.touch(*args, **kwargs)

    async def mset(self, *args, **kwargs):
        return await self.client.mset(*args, **kwargs)

    async def mget(self, *args, **kwargs):
        return await self.client.mget(*args, **kwargs)

    async def persist(self, *args, **kwargs) -> bool:
        return await self.client.persist(*args, **kwargs)

    async def expire(self, *args, **kwargs) -> bool:
        return await self.client.expire(*args, **kwargs)

    async def expire_at(self, *args, **kwargs) -> bool:
        return await self.client.expire_at(*args, **kwargs)

    async def pexpire(self, *args, **kwargs) -> bool:
        return await self.client.pexpire(*args, **kwargs)

    async def pexpire_at(self, *args, **kwargs) -> bool:
        return await self.client.pexpire_at(*args, **kwargs)

    async def get_lock(self, *args, **kwargs) -> "ALock":
        return await self.client.get_lock(*args, **kwargs)

    lock = get_lock

    async def delete_pattern(self, *args, **kwargs) -> int:
        kwargs.setdefault("itersize", self._default_scan_itersize)
        return await self.client.adelete_pattern(*args, **kwargs)

    async def ttl(self, *args, **kwargs) -> int:
        return await self.client.ttl(*args, **kwargs)

    async def pttl(self, *args, **kwargs) -> int:
        return await self.client.pttl(*args, **kwargs)

    async def iter_keys(self, *args, **kwargs) -> AsyncGenerator[Any, None]:
        async with contextlib.aclosing(self.client.iter_keys(*args, **kwargs)) as it:
            async for key in it:
                yield key

    async def keys(self, *args, **kwargs) -> list[Any]:
        return await self.client.keys(*args, **kwargs)

    async def sadd(self, *args, **kwargs) -> int:
        return await self.client.sadd(*args, **kwargs)

    async def scard(self, *args, **kwargs) -> int:
        return await self.client.scard(*args, **kwargs)

    async def sdiff(self, *args, **kwargs) -> builtins.set[Any]:
        return await self.client.sdiff(*args, **kwargs)

    async def sdiffstore(self, *args, **kwargs) -> int:
        return await self.client.sdiffstore(*args, **kwargs)

    async def sinter(self, *args, **kwargs) -> builtins.set[Any]:
        return await self.client.sinter(*args, **kwargs)

    async def sinterstore(self, *args, **kwargs) -> int:
        return await self.client.sinterstore(*args, **kwargs)

    async def smismember(self, *args, **kwargs) -> list[bool]:
        return await self.client.smismember(*args, **kwargs)

    async def sismember(self, *args, **kwargs) -> bool:
        return await self.client.sismember(*args, **kwargs)

    async def smembers(self, *args, **kwargs) -> builtins.set[Any]:
        return await self.client.smembers(*args, **kwargs)

    async def smove(self, *args, **kwargs) -> bool:
        return await self.client.smove(*args, **kwargs)

    async def spop(self, *args, **kwargs) -> builtins.set | Any:
        return await self.client.spop(*args, **kwargs)

    async def srandmember(self, *args, **kwargs) -> list | Any:
        return await self.client.srandmember(*args, **kwargs)

    async def srem(self, *args, **kwargs) -> int:
        return await self.client.srem(*args, **kwargs)

    async def sscan(self, *args, **kwargs) -> builtins.set[Any]:
        return await self.client.sscan(*args, **kwargs)

    async def sscan_iter(self, *args, **kwargs) -> AsyncGenerator[Any, None]:
        async with contextlib.aclosing(self.client.sscan_iter(*args, **kwargs)) as it:
            async for key in it:
                yield key

    async def sunion(self, *args, **kwargs) -> builtins.set[Any]:
        return await self.client.sunion(*args, **kwargs)

    async def sunionstore(self, *args, **kwargs) -> int:
        return await self.client.sunionstore(*args, **kwargs)

    async def hset(self, *args, **kwargs) -> int:
        return await self.client.hset(*args, **kwargs)

    async def hdel(self, *args, **kwargs) -> int:
        return await self.client.hdel(*args, **kwargs)

    async def hlen(self, *args, **kwargs) -> int:
        return await self.client.hlen(*args, **kwargs)

    async def hkeys(self, *args, **kwargs) -> list[Any]:
        return await self.client.hkeys(*args, **kwargs)

    async def hexists(self, *args, **kwargs) -> bool:
        return await self.client.hexists(*args, **kwargs)


# temp fix for django's #36047
# TODO: remove this when it's fixed in django
from django.core import signals  # noqa: E402
from django.core.cache import caches, close_caches  # noqa: E402


async def close_async_caches(**kwargs):
    for conn in caches.all(initialized_only=True):
        if getattr(conn, "is_async", False):
            await conn.aclose()
        else:
            conn.close()


signals.request_finished.connect(close_async_caches)
signals.request_finished.disconnect(close_caches)



import base64
import string
from datetime import timedelta
from typing import Optional, Type

import pytest

from django.conf import settings
from django.contrib.sessions.backends.base import SessionBase
from django.contrib.sessions.backends.cache import SessionStore as CacheSession
from django.core.cache import caches
from django.test import override_settings
from django.utils import timezone

from django_valkey.cache import ValkeyCache

SessionType = Type[SessionBase]


# Copied from Django's sessions test suite. Keep in sync with upstream.
# https://github.com/django/django/blob/main/tests/sessions_tests/tests.py
class SessionTestsMixin:
    # This does not inherit from TestCase to avoid any tests being run with this
    # class, which wouldn't work, and to allow different TestCase subclasses to
    # be used.

    backend: Optional[SessionType] = None  # subclasses must specify

    @pytest.fixture(autouse=True)
    def setup(self):
        self.session = self.backend()
        yield
        self.session.delete()

    def test_new_session(self):
        assert self.session.modified is False
        assert self.session.accessed is False

    def test_get_empty(self):
        assert self.session.get("cat") is None

    def test_store(self):
        self.session["cat"] = "dog"
        assert self.session.modified is True
        assert self.session.pop("cat") == "dog"

    def test_pop(self):
        self.session["some key"] = "exists"
        # Need to reset these to pretend we haven't accessed it:
        self.accessed = False
        self.modified = False

        assert self.session.pop("some key") == "exists"
        assert self.session.accessed is True
        assert self.session.modified is True
        assert self.session.get("some key") is None

    def test_pop_default(self):
        assert self.session.pop("some key", "does not exist") == "does not exist"
        assert self.session.accessed is True
        assert self.session.modified is False

    def test_pop_default_named_argument(self):
        assert (
            self.session.pop("some key", default="does not exist") == "does not exist"
        )
        assert self.session.accessed is True
        assert self.session.modified is False

    def test_pop_no_default_keyerror_raised(self):
        with pytest.raises(KeyError):
            self.session.pop("some key")

    def test_setdefault(self):
        assert self.session.setdefault("foo", "bar") == "bar"
        assert self.session.setdefault("foo", "baz") == "bar"
        assert self.session.accessed is True
        assert self.session.modified is True

    def test_update(self):
        self.session.update({"update key": 1})
        assert self.session.accessed is True
        assert self.session.modified is True
        assert self.session.get("update key") == 1

    def test_has_key(self):
        self.session["some key"] = 1
        self.session.modified = False
        self.session.accessed = False

        assert "some key" in self.session
        assert self.session.accessed is True
        assert self.session.modified is False

    def test_values(self):
        assert list(self.session.values()) == []
        assert self.session.accessed is True
        self.session["some key"] = 1
        self.session.modified = False
        self.session.accessed = False
        assert list(self.session.values()) == [1]
        assert self.session.accessed is True
        assert self.session.modified is False

    def test_keys(self):
        self.session["x"] = 1
        self.session.modified = False
        self.session.accessed = False
        assert list(self.session.keys()) == ["x"]
        assert self.session.accessed is True
        assert self.session.modified is False

    def test_items(self):
        self.session["x"] = 1
        self.session.modified = False
        self.session.accessed = False
        assert list(self.session.items()) == [("x", 1)]
        assert self.session.accessed is True
        assert self.session.modified is False

    def test_clear(self):
        self.session["x"] = 1
        self.session.modified = False
        self.session.accessed = False
        assert list(self.session.items()) == [("x", 1)]
        self.session.clear()
        assert list(self.session.items()) == []
        assert self.session.accessed is True
        assert self.session.modified is True

    def test_save(self):
        self.session.save()
        assert self.session.exists(self.session.session_key)

    def test_delete(self):
        self.session.save()
        self.session.delete(self.session.session_key)
        assert not self.session.exists(self.session.session_key)

    def test_flush(self):
        self.session["foo"] = "bar"
        self.session.save()
        prev_key = self.session.session_key
        self.session.flush()
        assert not self.session.exists(prev_key)
        assert self.session.session_key != prev_key
        assert self.session.session_key is None
        assert self.session.accessed is True
        assert self.session.modified is True

    def test_cycle(self):
        self.session["a"], self.session["b"] = "c", "d"
        self.session.save()
        prev_key = self.session.session_key
        prev_data = list(self.session.items())
        self.session.cycle_key()
        assert not self.session.exists(prev_key)
        assert self.session.session_key != prev_key
        assert list(self.session.items()) == prev_data

    def test_cycle_with_no_session_cache(self):
        self.session["a"], self.session["b"] = "c", "d"
        self.session.save()
        prev_data = self.session.items()
        self.session = self.backend(self.session.session_key)

        assert not hasattr(self.session, "_session_cache")
        self.session.cycle_key()
        assert self.session.items() == prev_data

    def test_save_doesnt_clear_data(self):
        self.session["a"] = "b"
        self.session.save()
        assert self.session["a"] == "b"

    def test_invalid_key(self):
        # Submitting an invalid session key (either by guessing, or if the db has
        # removed the key) results in a new key being generated.
        try:
            session = self.backend("1")
            session.save()
            assert session.session_key != "1"
            assert not self.session.get("cat")
            session.delete()
        finally:
            # Some backends leave a stale cache entry for the invalid
            # session key; make sure that entry is manually deleted
            session.delete("1")

    def test_session_key_empty_string_invalid(self):
        """Falsey values (Such as an empty string) are rejected."""
        self.session._session_key = ""
        assert not (self.session.session_key)

    def test_session_key_too_short_invalid(self):
        """Strings shorter than 8 characters are rejected."""
        self.session._session_key = "1234567"
        assert not (self.session.session_key)

    def test_session_key_valid_string_saved(self):
        """Strings of length 8 and up are accepted and stored."""
        self.session._session_key = "12345678"
        assert self.session.session_key == "12345678"

    def test_session_key_is_read_only(self):
        def set_session_key(session):
            session.session_key = session._get_new_session_key()

        with pytest.raises(AttributeError):
            set_session_key(self.session)

    # Custom session expiry
    def test_default_expiry(self):
        # A normal session has a max age equal to settings
        assert self.session.get_expiry_age() == settings.SESSION_COOKIE_AGE

        # So does a custom session with an idle expiration time of 0 (but it'll
        # expire at browser close)
        self.session.set_expiry(0)
        assert self.session.get_expiry_age() == settings.SESSION_COOKIE_AGE

    def test_custom_expiry_seconds(self):
        modification = timezone.now()

        self.session.set_expiry(10)

        date = self.session.get_expiry_date(modification=modification)
        assert date == modification + timedelta(seconds=10)

        age = self.session.get_expiry_age(modification=modification)
        assert age == 10

    def test_custom_expiry_timedelta(self):
        modification = timezone.now()

        # Mock timezone.now, because set_expiry calls it on this code path.
        original_now = timezone.now
        try:
            timezone.now = lambda: modification
            self.session.set_expiry(timedelta(seconds=10))
        finally:
            timezone.now = original_now

        date = self.session.get_expiry_date(modification=modification)
        assert date == modification + timedelta(seconds=10)

        age = self.session.get_expiry_age(modification=modification)
        assert age == 10

    def test_custom_expiry_datetime(self):
        modification = timezone.now()

        self.session.set_expiry(modification + timedelta(seconds=10))

        date = self.session.get_expiry_date(modification=modification)
        assert date == modification + timedelta(seconds=10)

        age = self.session.get_expiry_age(modification=modification)
        assert age == 10

    def test_custom_expiry_reset(self):
        self.session.set_expiry(None)
        self.session.set_expiry(10)
        self.session.set_expiry(None)
        assert self.session.get_expiry_age() == settings.SESSION_COOKIE_AGE

    def test_get_expire_at_browser_close(self):
        # Tests get_expire_at_browser_close with different settings and different
        # set_expiry calls
        with override_settings(SESSION_EXPIRE_AT_BROWSER_CLOSE=False):
            self.session.set_expiry(10)
            assert self.session.get_expire_at_browser_close() is False

            self.session.set_expiry(0)
            assert self.session.get_expire_at_browser_close() is True

            self.session.set_expiry(None)
            assert self.session.get_expire_at_browser_close() is False

        with override_settings(SESSION_EXPIRE_AT_BROWSER_CLOSE=True):
            self.session.set_expiry(10)
            assert self.session.get_expire_at_browser_close() is False

            self.session.set_expiry(0)
            assert self.session.get_expire_at_browser_close() is True

            self.session.set_expiry(None)
            assert self.session.get_expire_at_browser_close() is True

    def test_decode(self):
        # Ensure we can decode what we encode
        data = {"a test key": "a test value"}
        encoded = self.session.encode(data)
        assert self.session.decode(encoded) == data

    def test_decode_failure_logged_to_security(self, caplog):
        bad_encode = base64.b64encode(b"flaskdj:alkdjf").decode("ascii")
        with caplog.at_level("WARNING", "django.security.SuspiciousSession"):
            assert {} == self.session.decode(bad_encode)
        # The failed decode is logged.
        assert "corrupted" in caplog.records[0].message

    def test_actual_expiry(self):
        self.session = self.backend()  # reinitialize after overriding settings

        # Regression test for #19200
        old_session_key = None
        new_session_key = None
        try:
            self.session["foo"] = "bar"
            self.session.set_expiry(-timedelta(seconds=10))
            self.session.save()
            old_session_key = self.session.session_key
            # With an expiry date in the past, the session expires instantly.
            new_session = self.backend(self.session.session_key)
            new_session_key = new_session.session_key
            assert "foo" not in new_session
        finally:
            self.session.delete(old_session_key)
            self.session.delete(new_session_key)

    def test_session_load_does_not_create_record(self):
        """
        Loading an unknown session key does not create a session record.
        Creating session records on load is a DOS vulnerability.
        """
        session = self.backend("someunknownkey")
        session.load()

        assert session.session_key is None
        assert not session.exists(session.session_key)
        # provided unknown key was cycled, not reused
        assert session.session_key != "someunknownkey"

    def test_session_save_does_not_resurrect_session_logged_out_in_other_context(self):
        """
        Sessions shouldn't be resurrected by a concurrent request.
        """
        from django.contrib.sessions.backends.base import UpdateError

        # Create new session.
        s1 = self.backend()
        s1["test_data"] = "value1"
        s1.save(must_create=True)

        # Logout in another context.
        s2 = self.backend(s1.session_key)
        s2.delete()

        # Modify session in first context.
        s1["test_data"] = "value2"
        with pytest.raises(UpdateError):
            # This should throw an exception as the session is deleted, not
            # resurrect the session.
            s1.save()

        assert s1.load() == {}


class TestSession(SessionTestsMixin):
    backend = CacheSession

    def test_load_overlong_key(self):
        self.session._session_key = (string.ascii_letters + string.digits) * 20
        assert self.session.load() == {}

    def test_default_cache(self):
        self.session.save()
        assert caches["default"].get(self.session.cache_key) is not None

    @pytest.mark.skipif(
        caches["default"].client is not ValkeyCache,
        reason="settings is set for normal server",
    )
    def test_non_default_cache(self, settings):
        settings.CACHES = {
            "default": {
                "BACKEND": "django.core.cache.backends.dummy.DummyCache",
            },
            "sessions": {
                "BACKEND": "django_valkey.cache.ValkeyCache",
                "LOCATION": "valkey://localhost:6379",
            },
        }
        settings.SESSION_CACHE_ALIAS = "sessions"
        # Re-initialize the session backend to make use of overridden settings.
        self.session = self.backend()

        self.session.save()
        assert caches["default"].get(self.session.cache_key) is None
        assert caches["sessions"].get(self.session.cache_key) is not None

    def test_create_and_save(self):
        self.session = self.backend()
        self.session.create()
        self.session.save()
        assert caches["default"].get(self.session.cache_key) is not None



import copy
from collections.abc import Iterable
from typing import cast

import pytest
from pytest_django.fixtures import SettingsWrapper

from asgiref.compatibility import iscoroutinefunction
from django.core.cache import cache as default_cache, caches

from django_valkey.base import BaseValkeyCache
from django_valkey.cache import ValkeyCache


pytestmark = pytest.mark.anyio

if iscoroutinefunction(default_cache.clear):

    @pytest.fixture(scope="function")
    async def cache():
        yield default_cache
        await default_cache.aclear()

else:

    @pytest.fixture
    def cache() -> Iterable[BaseValkeyCache]:
        yield default_cache
        default_cache.clear()


@pytest.fixture
def key_prefix_cache(
    cache: ValkeyCache, settings: SettingsWrapper
) -> Iterable[ValkeyCache]:
    caches_setting = copy.deepcopy(settings.CACHES)
    caches_setting["default"]["KEY_PREFIX"] = "*"
    settings.CACHES = caches_setting
    yield cache


@pytest.fixture
def with_prefix_cache() -> Iterable[ValkeyCache]:
    with_prefix = cast(ValkeyCache, caches["with_prefix"])
    yield with_prefix
    with_prefix.clear()



import contextlib
import copy
from collections.abc import Iterable
from typing import cast

import pytest
from pytest import LogCaptureFixture
from pytest_django.fixtures import SettingsWrapper

from django.core.cache import caches, cache as default_cache

from valkey.exceptions import ConnectionError

from django_valkey.async_cache.cache import AsyncValkeyCache
from django_valkey.async_cache.client import AsyncHerdClient, AsyncDefaultClient

pytestmark = pytest.mark.anyio

methods_with_no_parameters = {"clear", "close"}

methods_with_one_required_parameters = {
    "decr",
    "delete",
    "delete_many",
    "delete_pattern",
    "get",
    "get_many",
    "has_key",
    "hkeys",
    "hlen",
    "incr",
    "incr_version",
    "decr_version",
    "keys",
    "lock",
    "mget",
    "pttl",
    "persist",
    "scard",
    "sdiff",
    "sinter",
    "smembers",
    "spop",
    "srandmember",
    "sscan",
    "sunion",
    "touch",
    "ttl",
}
methods_with_two_required_parameters = {
    "add",
    "expire",
    "expire_at",
    "hdel",
    "hexists",
    "pexpire",
    "pexpire_at",
    "sadd",
    "sdiffstore",
    "set",
    "sinterstore",
    "sismember",
    "smismember",
    "srem",
    "sunionstore",
}
methods_with_three_required_parameters = {"hset", "smove"}
methods_taking_dictionary = {"mset", "set_many"}
iter_methods = {
    "iter_keys",
    "sscan_iter",
}

no_herd_method = {
    "incr",
    "decr",
}


# TODO: when django adjusts the signal, remove this decorator (and the ones below)
@pytest.mark.filterwarnings("ignore:coroutine 'AsyncBackendCommands.close'")
class TestDjangoValkeyOmitException:
    @pytest.fixture
    async def conf_cache(self, settings: SettingsWrapper):
        caches_settings = copy.deepcopy(settings.CACHES)
        # NOTE: this files raises RuntimeWarning because `conn.close` was not awaited,
        # this is expected because django calls the signal manually during this test
        # to debug, put a `raise` in django.utils.connection.BaseConnectionHandler.close_all
        settings.CACHES = caches_settings
        return caches_settings

    @pytest.fixture
    async def conf_cache_to_ignore_exception(
        self, settings: SettingsWrapper, conf_cache
    ):
        conf_cache["doesnotexist"]["OPTIONS"]["IGNORE_EXCEPTIONS"] = True
        conf_cache["doesnotexist"]["OPTIONS"]["LOG_IGNORE_EXCEPTIONS"] = True
        settings.DJANGO_VALKEY_IGNORE_EXCEPTIONS = True
        settings.DJANGO_VALKEY_LOG_IGNORE_EXCEPTIONS = True

    @pytest.fixture
    async def ignore_exceptions_cache(
        self, conf_cache_to_ignore_exception
    ) -> AsyncValkeyCache:
        return cast(AsyncValkeyCache, caches["doesnotexist"])

    async def test_methods_with_no_argument_omit_exception(
        self, ignore_exceptions_cache: AsyncValkeyCache, subtests
    ):
        for m in methods_with_no_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                await method()

    async def test_methods_with_one_argument_omit_exception(
        self, ignore_exceptions_cache: AsyncValkeyCache, subtests
    ):
        for m in methods_with_one_required_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                if (
                    isinstance(default_cache.client, AsyncHerdClient)
                    and m in no_herd_method
                ):
                    pytest.skip(f"herd client doesn't support {m}")
                await method("abc")

    async def test_methods_with_two_argument_omit_exception(
        self, ignore_exceptions_cache: AsyncValkeyCache, subtests
    ):
        for m in methods_with_two_required_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                await method("abc", 1)

    @pytest.mark.skipif(
        not isinstance(default_cache.client, AsyncDefaultClient),
        reason="not supported by none default clients",
    )
    async def test_methods_with_three_argument_omit_exception(
        self, ignore_exceptions_cache: AsyncValkeyCache, subtests
    ):
        for m in methods_with_three_required_parameters:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                await method("abc", "def", "ghi")

    async def test_methods_taking_dictionary(
        self, ignore_exceptions_cache: AsyncValkeyCache, subtests
    ):
        for m in methods_taking_dictionary:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                await method({"abc": "def"})

    async def test_iterator_methods(
        self, ignore_exceptions_cache: AsyncValkeyCache, subtests
    ):
        for m in iter_methods:
            method = getattr(ignore_exceptions_cache, m)
            with subtests.test(method=method):
                async for i in method("abc"):
                    assert i is None

    async def test_get_django_omit_exceptions_many_returns_default_arg(
        self,
        ignore_exceptions_cache: AsyncValkeyCache,
    ):
        assert ignore_exceptions_cache._ignore_exceptions is True
        assert await ignore_exceptions_cache.aget_many(["key1", "key2", "key3"]) == {}

    async def test_get_django_omit_exceptions(
        self, caplog: LogCaptureFixture, ignore_exceptions_cache: AsyncValkeyCache
    ):
        assert ignore_exceptions_cache._ignore_exceptions is True
        assert ignore_exceptions_cache._log_ignored_exceptions is True

        assert await ignore_exceptions_cache.aget("key") is None
        assert await ignore_exceptions_cache.aget("key", "default") == "default"
        assert await ignore_exceptions_cache.aget("key", default="default") == "default"

        assert len(caplog.records) == 3
        assert all(
            record.levelname == "ERROR" and record.msg == "Exception ignored"
            for record in caplog.records
        )

    async def test_get_django_omit_exceptions_priority_1(
        self, settings: SettingsWrapper
    ):
        caches_setting = copy.deepcopy(settings.CACHES)
        caches_setting["doesnotexist"]["OPTIONS"]["IGNORE_EXCEPTIONS"] = True
        settings.CACHES = caches_setting
        settings.DJANGO_VALKEY_IGNORE_EXCEPTIONS = False
        cache = cast(AsyncValkeyCache, caches["doesnotexist"])
        assert cache._ignore_exceptions is True
        assert await cache.aget("key") is None

    async def test_get_django_omit_exceptions_priority_2(
        self, settings: SettingsWrapper
    ):
        caches_setting = copy.deepcopy(settings.CACHES)
        caches_setting["doesnotexist"]["OPTIONS"]["IGNORE_EXCEPTIONS"] = False
        settings.CACHES = caches_setting
        settings.DJANGO_VALKEY_IGNORE_EXCEPTIONS = True
        cache = cast(AsyncValkeyCache, caches["doesnotexist"])
        assert cache._ignore_exceptions is False
        with pytest.raises(ConnectionError):
            await cache.aget("key")

    async def test_error_raised_when_ignore_is_not_set(self, conf_cache):
        cache = caches["doesnotexist"]
        assert cache._ignore_exceptions is False
        assert cache._log_ignored_exceptions is False
        with pytest.raises(ConnectionError):
            await cache.get("key")


@pytest.fixture
async def key_prefix_cache(
    cache: AsyncValkeyCache, settings: SettingsWrapper
) -> Iterable[AsyncValkeyCache]:
    caches_setting = copy.deepcopy(settings.CACHES)
    caches_setting["default"]["KEY_PREFIX"] = "*"
    settings.CACHES = caches_setting
    yield cache


@pytest.fixture
async def with_prefix_cache() -> Iterable[AsyncValkeyCache]:
    with_prefix = cast(AsyncValkeyCache, caches["with_prefix"])
    yield with_prefix
    await with_prefix.clear()


@pytest.mark.filterwarnings("ignore:coroutine 'AsyncBackendCommands.close'")
class TestDjangoValkeyCacheEscapePrefix:
    async def test_delete_pattern(
        self, key_prefix_cache: AsyncValkeyCache, with_prefix_cache: AsyncValkeyCache
    ):
        await key_prefix_cache.aset("a", "1")
        await with_prefix_cache.aset("b", "2")
        await key_prefix_cache.adelete_pattern("*")
        assert await key_prefix_cache.ahas_key("a") is False
        assert await with_prefix_cache.aget("b") == "2"

    async def test_iter_keys(
        self, key_prefix_cache: AsyncValkeyCache, with_prefix_cache: AsyncValkeyCache
    ):
        await key_prefix_cache.aset("a", "1")
        await with_prefix_cache.aset("b", "2")
        async with contextlib.aclosing(key_prefix_cache.aiter_keys("*")) as keys:
            test_list = [k async for k in keys]

        assert test_list == ["a"]

    async def test_keys(
        self, key_prefix_cache: AsyncValkeyCache, with_prefix_cache: AsyncValkeyCache
    ):
        await key_prefix_cache.aset("a", "1")
        await with_prefix_cache.aset("b", "2")
        keys = await key_prefix_cache.akeys("*")
        assert "a" in keys
        assert "b" not in keys


@pytest.mark.filterwarnings("ignore:coroutine 'AsyncBackendCommands.close'")
async def test_custom_key_function(cache: AsyncValkeyCache, settings: SettingsWrapper):
    caches_setting = copy.deepcopy(settings.CACHES)
    caches_setting["default"]["KEY_FUNCTION"] = "tests.test_cache_options.make_key"
    caches_setting["default"]["REVERSE_KEY_FUNCTION"] = (
        "tests.test_cache_options.reverse_key"
    )
    settings.CACHES = caches_setting

    for key in ["foo-aa", "foo-ab", "foo-bb", "foo-bc"]:
        await cache.aset(key, "foo")

    res = await cache.adelete_pattern("*foo-a*")
    assert bool(res) is True

    keys = await cache.akeys("foo*")
    assert set(keys) == {"foo-bb", "foo-bc"}
    # ensure our custom function was actually called
    client = await cache.client.get_client(write=False)
    assert {k.decode() for k in await client.keys("*")} == ({"#1#foo-bc", "#1#foo-bb"})



import copy

from pytest_django.fixtures import SettingsWrapper

from valkey import ValkeyCluster

from django_valkey.cluster_cache.cache import ClusterValkeyCache


class TestDjangoValkeyCacheEscapePrefix:
    def test_keys(
        self,
        key_prefix_cache: ClusterValkeyCache,  # noqa: F811
        with_prefix_cache: ClusterValkeyCache,  # noqa: F811
    ):
        key_prefix_cache.set("a", "1")
        with_prefix_cache.set("b", "2")
        keys = key_prefix_cache.keys("*", target_nodes=ValkeyCluster.ALL_NODES)
        assert "a" in keys
        assert "b" not in keys


def test_custom_key_function(cache: ClusterValkeyCache, settings: SettingsWrapper):
    caches_setting = copy.deepcopy(settings.CACHES)
    caches_setting["default"]["KEY_FUNCTION"] = "tests.test_cache_options.make_key"
    caches_setting["default"]["REVERSE_KEY_FUNCTION"] = (
        "tests.test_cache_options.reverse_key"
    )
    settings.CACHES = caches_setting

    for key in ["{foo}-aa", "{foo}-ab", "{foo}-bb", "{foo}-bc"]:
        cache.set(key, "foo")

    res = cache.delete_pattern("*{foo}-a*")
    assert bool(res) is True

    keys = cache.keys("{foo}*", target_nodes=ValkeyCluster.ALL_NODES)
    assert set(keys) == {"{foo}-bb", "{foo}-bc"}
    # ensure our custom function was actually called
    assert {
        k.decode()
        for k in cache.client.get_client(write=False).keys(
            "*", target_nodes=ValkeyCluster.ALL_NODES
        )
    } == ({"#1#{foo}-bc", "#1#{foo}-bb"})



import base64
import string
from datetime import timedelta
from typing import Optional, Type

import pytest

from django.conf import settings
from django.contrib.sessions.backends.base import SessionBase
from django.contrib.sessions.backends.cache import SessionStore as CacheSession
from django.core.cache import caches
from django.test import override_settings
from django.utils import timezone


pytestmark = pytest.mark.anyio

SessionType = Type[SessionBase]


# Copied from Django's sessions test suite. Keep in sync with upstream.
# https://github.com/django/django/blob/main/tests/sessions_tests/tests.py
class SessionTestsMixin:
    # This does not inherit from TestCase to avoid any tests being run with this
    # class, which wouldn't work, and to allow different TestCase subclasses to
    # be used.

    backend: Optional[SessionType] = None  # subclasses must specify

    @pytest.fixture(autouse=True)
    async def setup(self):
        self.session = self.backend()
        yield
        await self.session.adelete()

    def test_new_session(self):
        assert self.session.modified is False
        assert self.session.accessed is False

    async def test_get_empty(self):
        assert await self.session.aget("cat") is None

    async def test_store(self):
        await self.session.aset("cat", "dog")
        assert self.session.modified is True
        assert await self.session.apop("cat") == "dog"

    async def test_pop(self):
        await self.session.aset("some key", "exists")
        # Need to reset these to pretend we haven't accessed it:
        self.accessed = False
        self.modified = False

        assert await self.session.apop("some key") == "exists"
        assert self.session.accessed is True
        assert self.session.modified is True
        assert await self.session.aget("some key") is None

    async def test_pop_default(self):
        assert await self.session.apop("some key", "does not exist") == "does not exist"
        assert self.session.accessed is True
        assert self.session.modified is False

    async def test_pop_default_named_argument(self):
        assert (
            await self.session.apop("some key", default="does not exist")
            == "does not exist"
        )
        assert self.session.accessed is True
        assert self.session.modified is False

    async def test_pop_no_default_keyerror_raised(self):
        with pytest.raises(KeyError):
            await self.session.apop("some key")

    async def test_setdefault(self):
        assert await self.session.asetdefault("foo", "bar") == "bar"
        assert await self.session.asetdefault("foo", "baz") == "bar"
        assert self.session.accessed is True
        assert self.session.modified is True

    async def test_update(self):
        await self.session.aupdate({"update key": 1})
        assert self.session.accessed is True
        assert self.session.modified is True
        assert await self.session.aget("update key") == 1

    async def test_has_key(self):
        await self.session.aset("some key", 1)
        self.session.modified = False
        self.session.accessed = False

        assert await self.session.ahas_key("some key")
        assert self.session.accessed is True
        assert self.session.modified is False

    async def test_values(self):
        assert list(await self.session.avalues()) == []
        assert self.session.accessed is True
        await self.session.aset("some key", 1)
        self.session.modified = False
        self.session.accessed = False
        assert list(await self.session.avalues()) == [1]
        assert self.session.accessed is True
        assert self.session.modified is False

    async def test_keys(self):
        await self.session.aset("x", 1)
        self.session.modified = False
        self.session.accessed = False
        assert list(await self.session.akeys()) == ["x"]
        assert self.session.accessed is True
        assert self.session.modified is False

    async def test_items(self):
        await self.session.aset("x", 1)
        self.session.modified = False
        self.session.accessed = False
        assert list(await self.session.aitems()) == [("x", 1)]
        assert self.session.accessed is True
        assert self.session.modified is False

    async def test_clear(self):
        await self.session.aset("x", 1)
        self.session.modified = False
        self.session.accessed = False
        assert list(await self.session.aitems()) == [("x", 1)]
        self.session.clear()
        assert list(await self.session.aitems()) == []
        assert self.session.accessed is True
        assert self.session.modified is True

    async def test_save(self):
        await self.session.asave()
        assert await self.session.aexists(self.session.session_key)

    async def test_delete(self):
        await self.session.asave()
        await self.session.adelete(self.session.session_key)
        assert not await self.session.aexists(self.session.session_key)

    async def test_flush(self):
        await self.session.aset("foo", "bar")
        await self.session.asave()
        prev_key = self.session.session_key
        await self.session.aflush()
        assert not await self.session.aexists(prev_key)
        assert self.session.session_key != prev_key
        assert self.session.session_key is None
        assert self.session.accessed is True
        assert self.session.modified is True

    async def test_cycle(self):
        await self.session.aset("a", "c")
        await self.session.aset("b", "d")
        await self.session.asave()
        prev_key = self.session.session_key
        prev_data = list(await self.session.aitems())
        await self.session.acycle_key()
        assert not await self.session.aexists(prev_key)
        assert self.session.session_key != prev_key
        assert list(await self.session.aitems()) == prev_data

    async def test_cycle_with_no_session_cache(self):
        await self.session.aset("a", "c")
        await self.session.aset("b", "d")
        await self.session.asave()
        prev_data = await self.session.aitems()
        self.session = self.backend(self.session.session_key)

        assert not hasattr(self.session, "_session_cache")
        await self.session.acycle_key()
        assert await self.session.aitems() == prev_data

    async def test_save_doesnt_clear_data(self):
        await self.session.aset("a", "b")
        await self.session.asave()
        assert await self.session.aget("a") == "b"

    async def test_invalid_key(self):
        # Submitting an invalid session key (either by guessing, or if the db has
        # removed the key) results in a new key being generated.
        try:
            session = self.backend("1")
            await session.asave()
            assert session.session_key != "1"
            assert not await self.session.aget("cat")
            await session.adelete()
        finally:
            # Some backends leave a stale cache entry for the invalid
            # session key; make sure that entry is manually deleted
            await session.adelete("1")

    def test_session_key_empty_string_invalid(self):
        """Falsey values (Such as an empty string) are rejected."""
        self.session._session_key = ""
        assert not (self.session.session_key)

    def test_session_key_too_short_invalid(self):
        """Strings shorter than 8 characters are rejected."""
        self.session._session_key = "1234567"
        assert not (self.session.session_key)

    def test_session_key_valid_string_saved(self):
        """Strings of length 8 and up are accepted and stored."""
        self.session._session_key = "12345678"
        assert self.session.session_key == "12345678"

    async def test_session_key_is_read_only(self):
        async def set_session_key(session):
            session.session_key = await session._aget_new_session_key()

        with pytest.raises(AttributeError):
            await set_session_key(self.session)

    # Custom session expiry
    async def test_default_expiry(self):
        # A normal session has a max age equal to settings
        assert await self.session.aget_expiry_age() == settings.SESSION_COOKIE_AGE

        # So does a custom session with an idle expiration time of 0 (but it'll
        # expire at browser close)
        await self.session.aset_expiry(0)
        assert await self.session.aget_expiry_age() == settings.SESSION_COOKIE_AGE

    async def test_custom_expiry_seconds(self):
        modification = timezone.now()

        await self.session.aset_expiry(10)

        date = await self.session.aget_expiry_date(modification=modification)
        assert date == modification + timedelta(seconds=10)

        age = await self.session.aget_expiry_age(modification=modification)
        assert age == 10

    async def test_custom_expiry_timedelta(self):
        modification = timezone.now()

        # Mock timezone.now, because set_expiry calls it on this code path.
        original_now = timezone.now
        try:
            timezone.now = lambda: modification
            await self.session.aset_expiry(timedelta(seconds=10))
        finally:
            timezone.now = original_now

        date = await self.session.aget_expiry_date(modification=modification)
        assert date == modification + timedelta(seconds=10)

        age = await self.session.aget_expiry_age(modification=modification)
        assert age == 10

    async def test_custom_expiry_datetime(self):
        modification = timezone.now()

        await self.session.aset_expiry(modification + timedelta(seconds=10))

        date = await self.session.aget_expiry_date(modification=modification)
        assert date == modification + timedelta(seconds=10)

        age = await self.session.aget_expiry_age(modification=modification)
        assert age == 10

    async def test_custom_expiry_reset(self):
        await self.session.aset_expiry(None)
        await self.session.aset_expiry(10)
        await self.session.aset_expiry(None)
        assert await self.session.aget_expiry_age() == settings.SESSION_COOKIE_AGE

    async def test_get_expire_at_browser_close(self):
        # Tests get_expire_at_browser_close with different settings and different
        # set_expiry calls
        with override_settings(SESSION_EXPIRE_AT_BROWSER_CLOSE=False):
            await self.session.aset_expiry(10)
            assert await self.session.aget_expire_at_browser_close() is False

            await self.session.aset_expiry(0)
            assert await self.session.aget_expire_at_browser_close() is True

            await self.session.aset_expiry(None)
            assert await self.session.aget_expire_at_browser_close() is False

        with override_settings(SESSION_EXPIRE_AT_BROWSER_CLOSE=True):
            await self.session.aset_expiry(10)
            assert await self.session.aget_expire_at_browser_close() is False

            await self.session.aset_expiry(0)
            assert await self.session.aget_expire_at_browser_close() is True

            await self.session.aset_expiry(None)
            assert await self.session.aget_expire_at_browser_close() is True

    def test_decode(self):
        # Ensure we can decode what we encode
        data = {"a test key": "a test value"}
        encoded = self.session.encode(data)
        assert self.session.decode(encoded) == data

    def test_decode_failure_logged_to_security(self, caplog):
        bad_encode = base64.b64encode(b"flaskdj:alkdjf").decode("ascii")
        with caplog.at_level("WARNING", "django.security.SuspiciousSession"):
            assert {} == self.session.decode(bad_encode)
        # The failed decode is logged.
        assert "corrupted" in caplog.records[0].message

    async def test_actual_expiry(self):
        self.session = self.backend()  # reinitialize after overriding settings

        # Regression test for #19200
        old_session_key = None
        new_session_key = None
        try:
            await self.session.aset("foo", "bar")
            await self.session.aset_expiry(-timedelta(seconds=10))
            await self.session.asave()
            old_session_key = self.session.session_key
            # With an expiry date in the past, the session expires instantly.
            new_session = self.backend(self.session.session_key)
            new_session_key = new_session.session_key
            assert not await new_session.ahas_key("foo")
        finally:
            await self.session.adelete(old_session_key)
            await self.session.adelete(new_session_key)

    async def test_session_load_does_not_create_record(self):
        """
        Loading an unknown session key does not create a session record.
        Creating session records on load is a DOS vulnerability.
        """
        session = self.backend("someunknownkey")
        await session.aload()

        assert session.session_key is None
        assert not await session.aexists(session.session_key)
        # provided unknown key was cycled, not reused
        assert session.session_key != "someunknownkey"

    async def test_session_save_does_not_resurrect_session_logged_out_in_other_context(
        self,
    ):
        """
        Sessions shouldn't be resurrected by a concurrent request.
        """
        from django.contrib.sessions.backends.base import UpdateError

        # Create new session.
        s1 = self.backend()
        await s1.aset("test_data", "value1")
        await s1.asave(must_create=True)

        # Logout in another context.
        s2 = self.backend(s1.session_key)
        await s2.adelete()

        # Modify session in first context.
        await s1.aset("test_data", "value2")
        with pytest.raises(UpdateError):
            # This should throw an exception as the session is deleted, not
            # resurrect the session.
            await s1.asave()

        assert await s1.aload() == {}


class TestSession(SessionTestsMixin):
    backend = CacheSession

    async def test_load_overlong_key(self):
        self.session._session_key = (string.ascii_letters + string.digits) * 20
        assert await self.session.aload() == {}

    async def test_default_cache(self):
        await self.session.asave()
        assert await caches["default"].aget(self.session.cache_key) is not None

    @pytest.mark.filterwarnings("ignore:coroutine 'AsyncBackendCommands.close'")
    async def test_non_default_cache(self, settings):
        settings.CACHES = {
            "default": {
                "BACKEND": "django.core.cache.backends.dummy.DummyCache",
            },
            "sessions": {
                "BACKEND": "django_valkey.async_cache.cache.AsyncValkeyCache",
                "LOCATION": "valkey://localhost:6379",
            },
        }
        settings.SESSION_CACHE_ALIAS = "sessions"
        # Re-initialize the session backend to make use of overridden settings.
        self.session = self.backend()

        await self.session.asave()
        assert await caches["default"].aget(self.session.cache_key) is None
        assert await caches["sessions"].get(self.session.cache_key) is not None

    async def test_create_and_save(self):
        self.session = self.backend()
        await self.session.acreate()
        await self.session.asave()
        assert await caches["default"].get(self.session.cache_key) is not None



SECRET_KEY = "django_tests_secret_key"

CACHES = {
    "default": {
        "BACKEND": "django_valkey.cluster_cache.cache.ClusterValkeyCache",
        "LOCATION": ["valkey://127.0.0.1:7005", "valkey://127.0.0.1:7005"],
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.cluster_cache.client.DefaultClusterClient"
        },
    },
    "doesnotexist": {
        "BACKEND": "django_valkey.cluster_cache.cache.ClusterValkeyCache",
        "LOCATION": "valkey://127.0.0.1:56379?db=0",
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.cluster_cache.client.DefaultClusterClient"
        },
    },
    "sample": {
        "BACKEND": "django_valkey.cluster_cache.cache.ClusterValkeyCache",
        "LOCATION": "valkey://127.0.0.1:7005:0,valkey://127.0.0.1:7002:0",
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.cluster_cache.client.DefaultClusterClient"
        },
    },
    "with_prefix": {
        "BACKEND": "django_valkey.cluster_cache.cache.ClusterValkeyCache",
        "LOCATION": "valkey://127.0.0.1:7005?db=0",
        "OPTIONS": {
            "CLIENT_CLASS": "django_valkey.cluster_cache.client.DefaultClusterClient"
        },
        "KEY_PREFIX": "test-prefix",
    },
}

# Include `django.contrib.auth` and `django.contrib.contenttypes` for mypy /
# django-stubs.

# See:
# - https://github.com/typeddjango/django-stubs/issues/318
# - https://github.com/typeddjango/django-stubs/issues/534
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
]

USE_TZ = False



from collections.abc import Iterable
from typing import Any

from valkey import Valkey
from valkey.typing import KeyT, EncodableT

from django_valkey.base_client import (
    DEFAULT_TIMEOUT,
    Backend,
    HerdCommonMethods,
    _main_exceptions,
)
from django_valkey.client.default import DefaultClient
from django_valkey.exceptions import ConnectionInterrupted


class HerdClient(HerdCommonMethods, DefaultClient):
    def set(
        self,
        key: KeyT,
        value: EncodableT,
        timeout: int | None = DEFAULT_TIMEOUT,
        version: int | None = None,
        client: Valkey | None = None,
        nx: bool = False,
        xx: bool = False,
    ):
        if timeout is DEFAULT_TIMEOUT:
            timeout = self._backend.default_timeout

        if timeout is None or timeout <= 0:
            return super().set(
                key,
                value,
                timeout=timeout,
                version=version,
                client=client,
                nx=nx,
                xx=xx,
            )

        packed = self._pack(value, timeout)
        real_timeout = timeout + self._herd_timeout

        return super().set(
            key,
            packed,
            timeout=real_timeout,
            version=version,
            client=client,
            nx=nx,
            xx=xx,
        )

    def get(self, key, default=None, version=None, client=None):
        packed = super().get(key, default=default, version=version, client=client)
        val, refresh = self._unpack(packed)

        if refresh:
            return default

        return val

    def get_many(self, keys, version=None, client=None):
        client = self._get_client(write=False, client=client)

        if not keys:
            return {}

        recovered_data = {}

        new_keys = [self.make_key(key, version=version) for key in keys]
        map_keys = dict(zip(new_keys, keys))

        try:
            pipeline = client.pipeline()
            for key in new_keys:
                pipeline.get(key)
            results = pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(new_keys, results):
            if value is None:
                continue

            val, refresh = self._unpack(self.decode(value))
            recovered_data[map_keys[key]] = None if refresh else val

        return recovered_data

    def mget(
        self,
        keys: Iterable[KeyT],
        version: int | None = None,
        client: Backend | Any | None = None,
    ) -> dict:
        client = self._get_client(write=False, client=client)
        if not keys:
            return {}

        recovered_data = {}

        new_keys = [self.make_key(key, version=version) for key in keys]

        try:
            results = client.mget(new_keys)
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

        for key, value in zip(keys, results):
            if value is None:
                continue
            val, refresh = self._unpack(self.decode(value))
            recovered_data[key] = None if refresh else val
        return recovered_data

    def set_many(
        self, data, timeout=DEFAULT_TIMEOUT, version=None, client=None, herd=True
    ):
        """
        Set a bunch of values in the cache at once from a dict of key/value
        pairs. This is much more efficient than calling set() multiple times.

        If timeout is given, that timeout will be used for the key; otherwise
        the default cache timeout will be used.
        """
        client = self._get_client(write=True, client=client)

        set_function = self.set if herd else super().set

        try:
            pipeline = client.pipeline()
            for key, value in data.items():
                set_function(key, value, timeout, version=version, client=pipeline)
            pipeline.execute()
        except _main_exceptions as e:
            raise ConnectionInterrupted(connection=client) from e

    def incr(self, *args, **kwargs):
        raise NotImplementedError

    def decr(self, *args, **kwargs):
        raise NotImplementedError

    def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None, client=None):
        client = self._get_client(write=True, client=client)

        value = self.get(key, version=version, client=client)
        if value is None:
            return False

        self.set(key, value, timeout=timeout, version=version, client=client)
        return True



import pickle

import pytest

from django.core.exceptions import ImproperlyConfigured

from django_valkey.serializers.pickle import PickleSerializer


class TestPickleSerializer:
    def test_invalid_pickle_version_provided(self):
        with pytest.raises(
            ImproperlyConfigured, match="PICKLE_VERSION value must be an integer"
        ):
            PickleSerializer({"PICKLE_VERSION": "not-an-integer"})

    def test_setup_pickle_version_not_explicitly_specified(self):
        serializer = PickleSerializer({})
        assert serializer._pickle_version == pickle.DEFAULT_PROTOCOL

    def test_setup_pickle_version_too_high(self):
        with pytest.raises(
            ImproperlyConfigured,
            match=f"PICKLE_VERSION can't be higher than pickle.HIGHEST_PROTOCOL:"
            f" {pickle.HIGHEST_PROTOCOL}",
        ):
            PickleSerializer({"PICKLE_VERSION": pickle.HIGHEST_PROTOCOL + 1})
